from __future__ import absolute_import
import sys
import inspect

import tensorflow as tf
from tensorflow.python.training import moving_averages

from torch.nn import functional as F

from ..core.blocks import ProcessingLayer
from .. import backend as A
from six.moves import range
from six.moves import zip


class PoolingLayer(ProcessingLayer):
    """
    Pooling layer that supports a type to choose what pooling methods to use.

    It is deprecated. Use specialized layer, e.g. `MaxpoolingLayer` instead.
    """
    def __init__(self, ksize, strides, padding="VALID", type="max", **kwargs):
        """
        Args:
            type: str
                Use max or average pooling. 'max' for max pooling, and 'avg'
                for average pooling.
        """
        super(PoolingLayer, self).__init__(**kwargs)
        self.ksize = ksize
        self.strides = strides
        self.padding = padding
        self.type = type

    def _setup(self):
        self.log("Padding method {}.".format(self.padding), debug=True)
        self.log("Pooling method {}.".format(self.type), debug=True)

    def _forward(self, input):
        if self.type == "max":
            self._data = A.nn.max_pool(input,
                                       self.ksize,
                                       self.strides,
                                       self.padding)
        elif self.type == "avg":
            self._data = A.nn.avg_pool(input,
                                       self.ksize,
                                       self.strides,
                                       self.padding)
        else:
            raise Exception("Type `{}` pooling is not supported.".format(
                self.type))

        return self._data


class MemoryMaxPoolingLayer(ProcessingLayer):
    NAME = "MPool"

    def __init__(self, ksize, strides, padding="VALID", type="max", **kwargs):
        """
        Args:
            type: str
                Use max or average pooling. 'max' for max pooling, and 'avg'
                for average pooling.
        """
        super(MemoryMaxPoolingLayer, self).__init__(**kwargs)
        self.ksize = ksize
        self.strides = strides
        self.padding = padding
        self.keep_memory = False
        self.memory = None

    def _setup(self):
        self.log("Padding method {}.".format(self.padding), debug=True)

    def _forward(self, input):
        if self.keep_memory:
            if self.memory is None:
                self._data, self.memory = A.nn.max_pool(input,
                                                        self.ksize,
                                                        self.strides,
                                                        self.padding,
                                                        return_indices=True)
            else:
                shape = A.get_shape(input)
                mshape = A.get_shape(self.memory)
                if mshape[:-2] != shape[:-2]:
                    if mshape[1:-2] != shape[1:-2]:
                        # Only the first dimension is allowed to be different.
                        raise ValueError("Shape not supported.")

                    idxs = A.cat([self.memory] * shape[0])
                else:
                    idxs = self.memory

                tshape = shape[:-2]
                tshape.append(shape[-1] * shape[-2])
                d = A.reshape(input, tshape)

                mshape = A.get_shape(idxs)
                mtshape = mshape[:-2]
                mtshape.append(mshape[-1] * mshape[-2])
                m = A.reshape(idxs, mtshape)

                out = A.gather(d, dim=-1, index=m)
                self._data = A.reshape(out, mshape)

        else:
            self._data, self.memory = A.nn.max_pool(input,
                                                    self.ksize,
                                                    self.strides,
                                                    self.padding,
                                                    return_indices=True)


        return self._data

    def set_state(self, state):
        self.keep_memory = state

MPoolLayer = MemoryMaxPoolingLayer


class _PoolingLayer(ProcessingLayer):
    def __init__(self, ksize, strides, padding, **kwargs):
        super(_PoolingLayer, self).__init__(**kwargs)

        self.ksize = ksize
        self.strides = strides
        self.padding = padding


class MaxPoolingLayer(_PoolingLayer):
    NAME = "MaxPool2D"

    def __init__(self, get_argmax_idx=False, **kwargs):
        """
        Args:
        get_argmax_idx: Bool
                Compute the indices where the max has been obtained or not. The
                indices is saved in `self.in_group_indices`.
        """
        super(MaxPoolingLayer, self).__init__(**kwargs)

        ksize = self.ksize
        t = type(ksize)
        if t is int:
            self.ksize = [ksize, ksize]
        elif t is list or t is tuple:
            if len(ksize) == 2:
                self.ksize = ksize
            else:
                raise ValueError("Only 2D pooling is supported. Gotten {}".format(ksize))
        else:
            raise ValueError("Value not understood".format(ksize))


        self.get_argmax_idx = get_argmax_idx

    def _forward(self, X_in):
        self.log("Padding method {}.".format(self.padding), debug=True)

        if type(X_in) is list:
            ret = []
            for X in X_in:
                ret.append(self._max_pooling(X))

            if self.get_argmax_idx:
                ret = list(zip(*ret))
                self._data, self.in_group_indices = ret[0], ret[1]
            else:
                self._data = ret
        else:
            ret = self._max_pooling(X_in)
            if self.get_argmax_idx:
                self._data, self.in_group_indices = ret
            else:
                self._data = ret

        return self._data

    def _max_pooling(self, X_in):
        if self.get_argmax_idx:
            return A.nn.max_pool_with_argmax(
                X_in, self.ksize, self.strides, self.padding)
        else:
            return A.nn.max_pool(X_in,
                                 self.ksize,
                                 self.strides,
                                 self.padding)



    def backward(self, X_in):
        if not self.get_argmax_idx:
            raise Exception("Set `get_argmax_idx` to True to use backward.")

        self._data_g = A.nn.max_unpooling(X_in, self.in_group_indices, self.ksize)

        return self._data_g


class MaxPooling1DLayer(_PoolingLayer):
    NAME = "MaxPooling1D"

    def _forward(self, x):
        self._data = A.nn.max_pool1d(x, self.ksize, self.strides, self.padding)
        return self._data

MaxPooling1D = MaxPooling1DLayer

class ReLULayer(ProcessingLayer):
    NAME = "ReLU"
    def _forward(self, input):
        self._data = A.nn.relu(
            input,
            name='fmap' if self.summarize_output else None)
        return self._data

    def backward(self, X_in):
        assert hasattr(self, "_data") and self._data is not None,\
            "Call forward first prior to backward"

        sign = A.cast(self._data > 10e-6, dtype=A.float32)  # Weed out very small values
        # NOTE: This can also implemented using backward propagated gradient of
        # ReLU. Try changing it when things go slow.
        self._data_g = X_in * sign

        return self._data_g

ReLU = ReLULayer


class MemoryReLULayer(ProcessingLayer):
    """
    ReLU layer that is able to memorize the activation patterns. If
    `keep_memory` is set to True, the layer will memorize the activation if it
    does not have one yet. The memorized activation pattern will be used until
    `keep_memory` is set to False.
    """

    NAME = "MReLU"
    def __init__(self, keep_memory=False, *args, **kwargs):
        super(MemoryReLULayer, self).__init__(*args, **kwargs)

        self.keep_memory = keep_memory
        self.memory = None

    def _forward(self, input):
        if self.keep_memory:
            if self.memory is None:
                self.memory = A.cast(input > 0, A.float32)
        else:
            self.memory = A.cast(input > 0, A.float32)

        self._data = A.mul(self.memory, input, name='fmap' if self.summarize_output else None)

        return self._data

    def set_state(self, keep_memory):
        self.keep_memory = keep_memory


MReLU = MemoryReLULayer


class ColorizationReLULayer(ProcessingLayer):
    def __init__(self, wipe_negative=False, **kwargs):
        super(ColorizationReLULayer, self).__init__(**kwargs)
        self.wipe_negative = wipe_negative

    def _forward(self, X_in):
        """
        This layer takes two inputs, the convoluted feature map and the color
        map that is downsampled accordingly to match the shape of the feature
        map.
        """
        F, C = X_in[0], X_in[1]
        F = A.nn.relu(F)
        C_list = []
        for i in range(3):
            C_list.append(F * A.expand_dims(C[..., i], -1))
        out = A.concat(
            concat_dim=3,
            values=C_list)
        if self.wipe_negative:
            out = A.nn.relu(out)

        self._data = out

        return self._data


class SigmoidLayer(ProcessingLayer):
    NAME = "Sigmoid"

    def _forward(self, x):
        self._data = A.nn.sigmoid(x)
        return self._data

Sigmoid = SigmoidLayer


class LRNLayer(ProcessingLayer):
    def __init__(self,
                 depth_radius=4,
                 bias=1,
                 alpha=0.001 / 9.0,
                 beta=0.75,
                 **kwargs):
        super(LRNLayer, self).__init__(**kwargs)
        self.depth_radius = depth_radius
        self.bias = bias
        self.alpha = alpha
        self.beta = beta

    def _forward(self, input):
        self._data = tf.nn.lrn(input,
                               self.depth_radius,
                               self.bias,
                               self.alpha,
                               self.beta)


class SoftmaxNormalizationLayer(ProcessingLayer):
    # A default name for the tensor returned by the layer.
    NAME = "Softmax_Normalization"

    def __init__(self, use_temperature=False, group_size=4, **kwargs):
        super(SoftmaxNormalizationLayer, self).__init__(**kwargs)
        self.use_temperature = use_temperature
        self.group_size = group_size

    def _forward(self, input):
        shape = input.get_shape().as_list()
        data = tf.reshape(input, [-1, shape[-1]])
        if self.use_temperature:
            T = self._get_variable("T",
                                   [1],
                                   initializer=tf.constant_initializer(10.0))
            data /= T
        if shape[-1] % self.group_size is not 0:
            raise Exception("Group size {} should evenly divide output channel"
                            " number {}".format(self.group_size, shape[-1]))
        num_split = shape[-1] // self.group_size
        self.log("Feature maps of layer {} is divided into {} group".format(
            self.name, num_split))
        data_split = tf.split(axis=1, num_or_size_splits=num_split, value=data)
        data_split = list(data_split)
        for i in range(0, len(data_split)):
            data_split[i] = tf.nn.softmax(data_split[i])
        data = tf.concat(axis=1, values=data_split,)
        output = tf.reshape(data, shape, SoftmaxNormalizationLayer.NAME)

        self._data = output


class GroupProcessingLayer(ProcessingLayer):
    """
    A abstract layer that processes layer by group. This is a meta class
    (`_forward` is not implemented).

    Two modes are possible for this layer. The first is to divide the
    neurons of this layer evenly using `group_size`. The second mode the
    input should be a list of tensors. In this case, `group_size` is
    ignored, and each tensor in the list is taken as a group. In both case,
    only the last dimension of the tensor indexes group member, the other
    dimensions index groups.
    """
    def __init__(self, group_size=4, **kwargs):
        super(GroupProcessingLayer, self).__init__(**kwargs)
        self.group_size = group_size
        # Members to be filled during `_pre_setup`.
        self.output_shape = None
        self.num_group = None
        self.shape_rank = None

    def _pre_forward(self, input):
        super(GroupProcessingLayer, self)._pre_forward(input)

        if type(input) is list:
            # Get the shape for the final output tensor.
            last_dim = 0
            for t in input:
                shape = t.get_shape().as_list()
                last_dim += shape[-1]
            self.output_shape = input[0].get_shape().as_list()
            self.output_shape[-1] = last_dim
            self.rank = len(self.output_shape)
            # No work actually done. Just logging and gather some meta data.
            self.num_group = len(input)
            self.log("Number of groups: {}".format(self.num_group))
            group_size_list = [t.get_shape().as_list()[-1] for t in input]
            self.log("Group size of each group are {}".format(group_size_list))
        else:
            self.output_shape = input.get_shape().as_list()
            self.rank = len(self.output_shape)
            if self.output_shape[-1] % self.group_size is not 0:
                raise Exception("Group size {} should evenly divide output channel"
                                " number {}".format(self.group_size,
                                                    self.output_shape[-1]))
            out_channel_num = self.output_shape[-1]
            self.num_group = out_channel_num // self.group_size
            self.log("Feature maps of layer {} is divided into {} group".format(
                self.name, self.num_group))
            self.log("All groups have equal size {}.".format(self.group_size))


class GroupSoftmaxLayer(GroupProcessingLayer):
    # A default name for the tensor returned by the layer.
    NAME = "GSMax"

    def __init__(self, concat_output=True, use_temperature=False, **kwargs):
        """
        Args:
            concat_output: Boolean
                Whether to concat the scattered list into one tensor.
        """
        super(GroupSoftmaxLayer, self).__init__(**kwargs)
        self.use_temperature = use_temperature
        self.concat_output = concat_output

    def _forward(self, input):
        # Divide the input into list if not already.
        if type(input) is list:
            splitted_input = input
        else:
            out_channel_num = self.output_shape[-1]
            if self.num_group == out_channel_num:
                # Means the situation has degenerated into sigmoid activation
                # Just compute and return
                self._data = tf.nn.sigmoid(input)
                return

            splitted_input = tf.split(axis=self.rank-1, num_or_size_splits=self.num_group, value=input)
            splitted_input = list(splitted_input)

        # Add temperature if needed
        if self.use_temperature:
            for t in splitted_input:
                T = self._get_variable(
                    "T",
                    [1],
                    initializer=tf.constant_initializer(10.0))
                t /= T

        for i, t in enumerate(splitted_input):
            # Augment each split with a constant 1.
            ground_state_shape = self.output_shape[0:-1]
            ground_state_shape.append(1)
            self.ground_state = tf.constant(1.0, shape=ground_state_shape)

            # Compute group softmax.
            augmented_t = tf.concat(axis=self.rank-1, values=[t, self.ground_state])
            softmax_t = tf.nn.softmax(augmented_t)
            splitted_input[i] = softmax_t[..., 0:-1]

        output = splitted_input

        if self.concat_output:
            output = tf.concat(axis=self.rank-1, values=splitted_input)

        self._data = output


class CollapseOutLayer(GroupProcessingLayer):
    """
    `CollapseOutLayer` is to collapse the a subspace into a one-dimensional
    space. It could be Maxout or AverageOut. To ensure backward compatibility,
    by default, Maxout is the choice.

    It is not merged into PoolingLayer is because CollapseOutLayer should
    strictly use `VALID` padding so to decouple these two type of padding,
    these two layers are separated.
    """
    # A default name for the tensor returned by the layer.
    MAXOUT_NAME = "MaxOut"
    AVEOUT_NAME = "AverageOut"

    def __init__(self, type="maxout", **kwargs):
        super(CollapseOutLayer, self).__init__(**kwargs)
        self.type = type

    def _forward(self, input):
        if type(input) is list:
            # process each tensor once by one and combine them
            reduced_t_list = []
            for t in input:
                reduced_t = self._reduce(t)
                reduced_t_list.append(reduced_t)

            output = tf.stack(reduced_t_list, axis=-1)
        else:
            shape_by_group = self.output_shape[:]
            shape_by_group[-1] = self.num_group
            shape_by_group.append(self.group_size)
            buff_tensor = tf.reshape(input, shape_by_group)
            output = self._reduce(buff_tensor)

        self._data = output

    def _reduce(self, tensor):
        shape = tensor.get_shape().as_list()
        if self.type is "maxout":
            output = tf.reduce_max(tensor,
                                   axis=len(shape)-1,
                                   name=CollapseOutLayer.MAXOUT_NAME)
        elif self.type is "average_out":
            output = tf.reduce_mean(tensor,
                                    axis=len(shape)-1,
                                    name=CollapseOutLayer.AVEOUT_NAME)
        else:
            raise Exception("Type of `CollapseOutLayer` should be 'maxout' or"
                            "'average_out'! {} is given.".format(self.type))

        return output


class BatchNormalizationLayer(ProcessingLayer):
    NAME = "BN"

    def __init__(self,
                 channel_num,
                 beta_init=0,
                 gamma_init=1,
                 eps=1e-5,
                 momentum=0.1,
                 track_running_stats=True,
                 fix_gamma=False,
                 share_gamma=False,
                 dim_num=None,
                 **kwargs):
        """
        `beta_init`, `gamma_init`, `share_gamma` are only useful in tensorflow
        backend.
        """
        super(BatchNormalizationLayer, self).__init__(**kwargs)
        self.channel_num = channel_num
        self.fix_gamma = fix_gamma
        self.eps = eps
        self.momentum = momentum
        self.track_running_stats = track_running_stats

        # TODO: make the init setter work for torch as well.
        if A.backend() == A.TF:
            self.beta_init = float(beta_init)
            self.gamma_init = float(gamma_init)
            self.share_gamma = share_gamma

    def _setup(self):
        if A.backend() == A.TORCH:
            self.weights = self._get_variable("weights",
                                            self.channel_num,
                                            self._get_initializer({"name": "range_uniform", "low": 0, "high": 1}))
            self.biases = self._get_variable("biases",
                                           self.channel_num,
                                           self._get_initializer({"name": "constant", "value": 0}))

            if self.track_running_stats:
                # Running mean and variance that require no grad.
                self.running_mean = self._get_variable("running_mean",
                                                    self.channel_num,
                                                    self._get_initializer({"name": "constant", "value": 0}),
                                                    trainable=False)
                self.running_var = self._get_variable("running_var",
                                                    self.channel_num,
                                                    self._get_initializer({"name": "constant", "value": 1}),
                                                    trainable=False)
                self.num_batches_tracked = self._get_variable("num_batches_tracked",
                                                            0,
                                                            self._get_initializer({"name": "scalar", "value": 0}),
                                                            trainable=False)
            else:
                self.running_mean = None
                self.running_var = None
                self.num_batches_tracked = None

            return

        elif A.backend() == A.TF:
            # Logging.
            if self.gamma_init:
                self.log("Gamma initial value is {}.".format(self.gamma_init))
                if self.fix_gamma:
                    self.log("Gamma is fixed to during training.")
                else:
                    self.log("Gamma is trainable.")
            else:
                self.log("Gamma is not used during training.")

            self.beta, loss = self._variable_with_weight_decay(
                'beta',
                shape=[self.channel_num],
                init_para={"name": "constant", "value": self.beta_init})

            if loss is not None:
                self._loss = loss

            if self.fix_gamma:
                self.gamma = tf.constant(
                    self.gamma_init,
                    shape=[] if self.share_gamma else [self.channel_num],
                    name="gamma")
            else:
                self.gamma, loss = self._variable_with_weight_decay(
                    'gamma',
                    shape=[] if self.share_gamma else [self.channel_num],
                    init_para={"name": "constant", "value": self.gamma_init})

                if loss is not None:
                    self._loss += loss
        else:
            raise ValueError("Backend {} is not supported".format(A.backend()))


    def _forward(self, input):
        # TODO: the backend dependent code should be changed to a backend
        # independent one.
        if A.backend() == A.TF:
            self._data = A.nn.bn(input,
                                self.channel_num,
                                self.gamma,
                                self.beta,
                                self.is_val,
                                A.get_step(),
                                fix_gamma=self.fix_gamma,
                                share_gamma=self.share_gamma,
                                name="bn" if self.summarize_output else None)
        elif A.backend() == A.TORCH:
            exponential_average_factor = 0.0

            if not self.is_val and self.track_running_stats:
                if self.num_batches_tracked is not None:
                    self.num_batches_tracked += 1
                    if self.momentum is None:  # use cumulative moving average
                        exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                    else:  # use exponential moving average
                        exponential_average_factor = self.momentum

            self._data = A.nn.bn(input, self.channel_num,
                                 self.weights, self.biases,
                                 self.running_mean, self.running_var,
                                 self.is_val, self.track_running_stats,
                                 exponential_average_factor, self.eps,
                                 name="bn" if self.summarize_output else None)

        return self._data

    def backward(self, X_in):
        """
        Just pass the data back for now.
        """
        self._data = X_in

        return self._data

BN = BatchNormalizationLayer


class DropoutLayer(ProcessingLayer):
    NAME = "Dropout"

    def __init__(self, keep_prob=0.5, **kwargs):
        super(DropoutLayer, self).__init__(**kwargs)
        self.keep_prob = keep_prob

    def _forward(self, input):
        self._data = A.nn.dropout(input, self.keep_prob, self.is_val)

Dropout = DropoutLayer


class FixedWeightInnerProductLayer(ProcessingLayer):
    NAME = "FixedWeightIP"

    def __init__(self, weights, **kwargs):
        super(FixedWeightInnerProductLayer, self).__init__(**kwargs)
        self.weights = weights

    def _setup(self):
        self.weights = A.Tensor(self.weights)

    def _forward(self, x):
        self._data = A.nn.inner_product(x,
                                        self.weights,
                                        name='fmap' if self.summarize_output else None)

        return self._data

FixedWeightIPLayer = FixedWeightInnerProductLayer


class L2NormalizationLayer(ProcessingLayer):
    NAME = "L2Norm"

    def _forward(self, x):
        dim = len(x.shape)
        assert dim == 2, "dim {} is not supported".format(dim)

        norm = A.nn.l2_norm(x, 1)
        norm = A.expand_dims(norm, axis=1)
        self._data = x / norm

        return self._data

L2Norm = L2NormalizationLayer


__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



"""
This module contains ops to compute quantities related to matrices.
"""
import numpy as np
import torch as th
from tqdm import tqdm

from .. import backend as A
from ..utils import glog as log
from ..utils.decorators import coroutine
from ..core.events import DoneEvent


def lanczos_tridiagonalization_slow(H, q=None, K=None):
    """
    Given a symmetric matrix, factorize it as

    .. math::

        AQ = QT

    where :math:`Q` is orthogonal, and :math:`T` is tridiagonal. q would be the
    first column of Q. It is possible that rank(Q) is less than rank(H), since
    the algorithm may terminate when an invariant subspace is found. This could
    be an intermediate step to obtain the eigenvalues of H, since T is a matrix
    that is similar with H; it is also used (as an intermediate step) to
    approximate the eigenspectrum of H, when H is too large to compute its
    eigenspectrum exactly.

    More in *The Full Spectrum of Deep Net Hessians At Scale: Dynamics with
    Sample Size*, https://arxiv.org/abs/1811.07062.

    Args:
        H: a symmetric N X N matrix
        q: a N X 1 vector. If None, one will be created randomly.
        K: an integer. The number of iterations to run. If None, would run until
           an invariant subspace is found, or N iterations.

    Returns:
        eig: a tuple of two tensor: a K dimension vector contains eigenvalues
             of H; a K X N matrix contains eigenvectors.
    """
    N = A.get_shape(H)[0]

    if K is None:
        K = N

    if K > 10 ** 4:
        log.warning("Iteration number K {} may be too large".format(K))


    T = A.zeros((K, K))
    alpha = A.zeros(K)
    beta = A.zeros(K)
    V = A.zeros((N, K))

    k = 0
    beta_k = 1
    while beta_k != 0 and k < K:
        if k == 0:
            if q is None:
                # Sample a vector from N(0, I)
                v = A.randn((N,))
                v = v / A.nn.l2_norm(v)
                w = H @ v
            else:
                v = q
        else:
            w = H @ v - beta[k-1] * V[:, k-1]

        alpha_k = v @ w
        w = w - alpha_k * v
        w = w - V @ V.t() @ w
        beta_k = A.nn.l2_norm(w)

        V[:, k] = v
        alpha[k] = alpha_k
        beta[k] = beta_k

        v = w / beta_k

        k += 1

    # Construct T and compute its eigenvalues.
    diag_idxs = A.range(0, K)
    T[diag_idxs, diag_idxs] = alpha
    lower_diag_idxs = (diag_idxs + 1)[:-1]
    T[lower_diag_idxs, diag_idxs[:-1]] = beta[:-1]
    upper_diag_idxs = (diag_idxs - 1)[1:]
    T[upper_diag_idxs, diag_idxs[1:]] = beta[:-1]

    return A.symeig(T)


@coroutine
def lanczos_tridiagonalization_fast_cr(N, q=None, K=None, return_beta_k=False):
    """
    Coroutine version of the Lanczos tridiagonalization algorithm that
    delegates Hessian vector product. See `lanczos_tridiagonalization_fast`.
    """
    if K is None:
        K = N

    if K > 10 ** 4:
        log.warning("Iteration number K {} may be too large".format(K))


    T = A.zeros((K, K))
    alpha = A.zeros(K)
    beta = A.zeros(K)

    k = 0
    beta_k = 1

    pbar = tqdm(total=K)
    while beta_k != 0 and k < K:
        if k == 0:
            if q is None:
                # Sample a vector from N(0, I)
                v = A.randn((N,))
                # v = A.cast(v, th.float64) # TODO: fix here. A stub for adding
                # double precision support in the future.
                v = v / A.nn.l2_norm(v)
            else:
                v = q

            w = yield v # Delegate the computation of Hessian vector products up.
            # w = hessian_vector_product(H, v)
        else:
            Hv = yield v
            w = Hv - beta[k-1] * v_prev
            # w = hessian_vector_product(H, v) - beta[k-1] * v_prev

        alpha_k = v @ w
        w = w - alpha_k * v
        beta_k = A.nn.l2_norm(w)

        alpha[k] = alpha_k
        beta[k] = beta_k

        v_prev = v
        v = w / beta_k

        k += 1
        pbar.update(1)

    pbar.close()

    # Construct T and compute its eigenvalues.
    diag_idxs = A.range(0, K)
    T[diag_idxs, diag_idxs] = alpha
    lower_diag_idxs = (diag_idxs + 1)[:-1]
    T[lower_diag_idxs, diag_idxs[:-1]] = beta[:-1]
    upper_diag_idxs = (diag_idxs - 1)[1:]
    T[upper_diag_idxs, diag_idxs[1:]] = beta[:-1]

    eigenvalues, eigenvectors = A.symeig(T)

    if return_beta_k:
        yield eigenvalues, eigenvectors, beta[-1]
    else:
        yield eigenvalues, eigenvectors


def lanczos_tridiagonalization_fast(H, q=None, K=None, return_beta_k=False):
    """
    Lanczos algorithm that does not re-orthogonalize in each iteration. The
    time and space complexity is significantly smaller.

    The last off diagonal entry can be used to provide error bound on the
    eigenvalue estimation. In that, if an error estimation is needed, specify
    `return_beta_k` as True.

    See also `lanczos_tridiagonalization_slow`.
    """
    N = A.get_shape(H)[0]

    if K is None:
        K = N

    if K > 10 ** 4:
        log.warning("Iteration number K {} may be too large".format(K))


    T = A.zeros((K, K))
    alpha = A.zeros(K)
    beta = A.zeros(K)

    k = 0
    beta_k = 1
    while beta_k != 0 and k < K:
        if k == 0:
            if q is None:
                # Sample a vector from N(0, I)
                v = A.randn((N,))
                v = v / A.nn.l2_norm(v)
            else:
                v = q

            w = H @ v
        else:
            w = H @ v - beta[k-1] * v_prev

        alpha_k = v @ w
        w = w - alpha_k * v
        beta_k = A.nn.l2_norm(w)

        alpha[k] = alpha_k
        beta[k] = beta_k

        v_prev = v
        v = w / beta_k

        k += 1

    # Construct T and compute its eigenvalues.
    diag_idxs = A.range(0, K)
    T[diag_idxs, diag_idxs] = alpha
    lower_diag_idxs = (diag_idxs + 1)[:-1]
    T[lower_diag_idxs, diag_idxs[:-1]] = beta[:-1]
    upper_diag_idxs = (diag_idxs - 1)[1:]
    T[upper_diag_idxs, diag_idxs[1:]] = beta[:-1]

    eigenvalues, eigenvectors = A.symeig(T)

    if return_beta_k:
        return eigenvalues, eigenvectors, beta[-1]
    else:
        return eigenvalues, eigenvectors


@coroutine
def lanczos_spectrum_approx_cr(N, iter_num, K, n_vec):
    """
    The version of Lanczos spectrum approximation that delegates the computation
    of Hessian vector product. For the algorithm, see
    `lanczos_spectrum_approx`. Notably, in the coroutine version, the Hessian
    passed in the normal version is `N`, the dimension of the Hessian instead.
    """
    eigenvalues = A.zeros((n_vec, iter_num))
    eigenvectors = A.zeros((n_vec, iter_num, iter_num))
    print("Running Lanczos algorithm ... ")
    for i in tqdm(range(n_vec)):
        f, ret = lanczos_tridiagonalization_fast_cr(N, K=iter_num)
        Hv = yield ret
        while True:
            v = f.send(Hv)
            if type(v) is tuple:
                # We have reached the end of the iteration
                f.close()
                break
            Hv = yield v
        # The last returned values from coroutine below are the eigenvalues and
        # eigenvectors computed.
        eigenvalues[i], eigenvectors[i] = v

        # Essentially, the above code computes the following line in the normal version.
        # eigenvalues[i], eigenvectors[i] = lanczos_tridiagonalization_fast(H, K=iter_num)
        # eigenvalues[i], eigenvectors[i] = lanczos_tridiagonalization_slow(H, K=iter_num)

    t = A.linspace(-1, 1, K)
    psi = A.zeros(K)
    sigma = 2 / ((iter_num - 1) * ((8 * np.log(1.25)) ** (1/2)))
    gaussian = lambda t: 1 / (sigma * (2 * np.pi) ** (1/2)) * np.e ** (-(1/2) * (t / sigma) ** 2)
    print("Running eigenspectrum approximation ... ")
    tol = 1e-08
    width = sigma * np.sqrt(-2.0 * np.log(tol))
    for k in tqdm(range(K)):
        mass = 0
        for i in range(n_vec):
            for j in range(iter_num):
                if abs(t[k] - eigenvalues[i, j]) > width:
                    continue
                for v in eigenvectors[i, :, j]:
                    if abs(v) > 10e-7:
                        break
                mass += v ** 2 * gaussian(t[k] - eigenvalues[i, j])
        mass /= n_vec
        psi[k] = mass


    # return psi
    yield psi, DoneEvent("Lanczos Spectrum Approximation Done.")


def lanczos_spectrum_approx(H, iter_num, K, n_vec):
    """
    The algorithm that uses Lanczos algorithm to approximate the eigenspectrum
    of a symmetric matrix H. For more details, refer to *The Full Spectrum of
    Deep Net Hessians At Scale: Dynamics with Sample Size*,
    https://arxiv.org/abs/1811.07062.

    Args:
        H: a symmetric matrix of which the eigenspectrum is to be approximated,
           and is in the range [-1, 1].
        iter_num: the number of iteration to run in the Lanczos algorithm.
        K: the number of eigenvalues used as nodes (sampled points), around
           which the density of eigenvalues is approximated.
        n_vec: the number of unit-norm vectors to randomly sample to serve as
        the starting point of the Lanczos algorithm.

    Return:
        psi: a vector of dimension K. Density of the spectrum of H evaluated at
             K evenly distributed points in the range [-1, 1].
    """
    shape = A.get_shape(H)
    N = shape[0]
    eigenvalues = A.zeros((n_vec, iter_num))
    eigenvectors = A.zeros((n_vec, iter_num, iter_num))
    for i in range(n_vec):
        eigenvalues[i], eigenvectors[i] = lanczos_tridiagonalization_fast(H, K=iter_num)
        # eigenvalues[i], eigenvectors[i] = lanczos_tridiagonalization_slow(H, K=iter_num)

    t = A.linspace(-1, 1, K)
    psi = A.zeros(K)
    sigma = 2 / ((iter_num - 1) * ((8 * np.log(1.25)) ** (1/2)))
    gaussian = lambda t: 1 / (sigma * (2 * np.pi) ** (1/2)) * np.e ** (-(1/2) * (t / sigma) ** 2)
    for k in range(K):
        mass = 0
        for i in range(n_vec):
            for j in range(iter_num):
                for v in eigenvectors[i, :, j]:
                    if abs(v) > 10e-7:
                        break
                mass += v ** 2 * gaussian(t[k] - eigenvalues[i, j])
        mass /= n_vec
        psi[k] = mass

    return psi


@coroutine
def center_unit_eig_normalization_cr(N, iter_num, kappa):
    """
    Given a symmetric matrix, normalize the eigenspectrum of it to [-1, 1].

    Args:
        N: int
            The number of parameters.
        iter_num: int
            The number of interation to run in the underlyiong Lanzcos
            algorithm.
        kappa: float
            Margin percentage.

    Return:
        Parameters to denormalize the Hessian computed from the grad.
    """
    f, ret = lanczos_tridiagonalization_fast_cr(N, K=iter_num, return_beta_k=True)
    Hv = yield ret
    while True:
        v = f.send(Hv)
        if type(v) is tuple:
            # We have reached the end of the iteration
            f.close()
            break
        Hv = yield v
    # The last returned values from coroutine below are the eigenvalues and
    # eigenvectors computed.
    eigenvalues, eigenvectors, beta_k = v

    # eigenvalues, eigenvectors, beta_k = lanczos_tridiagonalization_fast(H, K=M, return_beta_k=True)
    # The following lines incorporate error bounds through Ritz
    # approximation. For the derivation, refer to p. 575, _Matrix Computation_,
    # also p. 318, _The Symmetric Eigenvalue Problem_.
    eig_min = eigenvalues[0] - abs(beta_k * eigenvectors[-1][0])
    eig_max = eigenvalues[-1] + abs(beta_k * eigenvectors[-1][-1])
    delta = kappa * (eig_max - eig_min)
    eig_min = eig_min - delta
    eig_max = eig_max + delta
    c = (eig_min + eig_max) / 2
    d = (eig_max - eig_min) / 2

    # We only return the parameters c, d computed, since grad would be
    # recomputed normally afterwards. The normalization should happen there.
    yield c, d


def center_unit_eig_normalization(H, M, kappa):
    """
    Given a symmetric matrix, normalize the eigenspectrum of it to [-1, 1].

    Args:
        M: int
            The number of interation to run in the underlyiong Lanzcos
            algorithm.
        kappa: float
            Margin percentage.

    Return:
        Normalized matrix, and parameters to denormalize it
    """
    N = A.get_shape(H)[0]
    eigenvalues, eigenvectors, beta_k = lanczos_tridiagonalization_fast(H, K=M, return_beta_k=True)
    # The following lines incorporate error bounds through Ritz
    # approximation. For the derivation, refer to p. 575, _Matrix Computation_,
    # also p. 318, _The Symmetric Eigenvalue Problem_.
    eig_min = eigenvalues[0] - abs(beta_k * eigenvectors[-1][0])
    eig_max = eigenvalues[-1] + abs(beta_k * eigenvectors[-1][-1])
    delta = kappa * (eig_max - eig_min)
    eig_min = eig_min - delta
    eig_max = eig_max + delta
    c = (eig_min + eig_max) / 2
    d = (eig_max - eig_min) / 2
    return (H - c * A.eye(N)) / d, c, d


def center_unit_eig_denormalization(H, c, d):
    """
    Given a matrix, and two parameters (obtained when doing normalization using
    'center_unit_eig_normalization`), denormalize the eigenspectrum of matrix
    `H` to its original range.
    """
    N = A.get_shape(H)[0]
    return H * d + c * A.eye(N)



from __future__ import absolute_import
from akid import GraphBrain
from akid.layers import (
    ConvolutionLayer,
    PoolingLayer,
    ReLULayer,
    InnerProductLayer,
    SoftmaxWithLossLayer,
    LRNLayer,
    BatchNormalizationLayer,
    DropoutLayer,
    MergeLayer,
    ReshapeLayer,
    PaddingLayer,
    CollapseOutLayer,
    GroupSoftmaxLayer,
    MaxPoolingLayer
)
from akid import backend as A
from six.moves import range


class AlexNet(GraphBrain):
    """
    A class for alex net specifically.
    """
    def __init__(self, in_channel_num, dataset="imagenet",  **kwargs):
        super(AlexNet, self).__init__(**kwargs)

        self.in_channel_num = in_channel_num

        if dataset == "imagenet":
            self._build_imagenet_model()
        else:
            self._build_cifar10_model()

    def _build_imagenet_model(self):
        wd={"type": "l2", "scale": 1e-4}
        init = {"name": "xavier"}
        self.attach(ConvolutionLayer(ksize=[11, 11],
                                     strides=[4, 4],
                                     padding="VALID",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=3,
                                     out_channel_num=64,
                                     name='conv1'))
        self.attach(ReLULayer(name='relu1'))
        self.attach(MaxPoolingLayer(ksize=[3, 3],
                                    strides=[2, 2],
                                    padding="VALID",
                                    name="pool1"))
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=64,
                                     out_channel_num=192,
                                     name='conv2'))
        self.attach(ReLULayer(name='relu2'))
        self.attach(MaxPoolingLayer(ksize=[3, 3],
                                    strides=[2, 2],
                                    padding="VALID",
                                    name="pool2"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=192,
                                     out_channel_num=384,
                                     name='conv3'))
        self.attach(ReLULayer(name='relu3'))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=384,
                                     out_channel_num=256,
                                     name='conv4'))
        self.attach(ReLULayer(name='relu4'))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=256,
                                     out_channel_num=256,
                                     name='conv5'))
        self.attach(ReLULayer(name='relu5'))
        self.attach(MaxPoolingLayer(ksize=[3, 3],
                                    strides=[2, 2],
                                    padding="VALID",
                                    name="pool2"))
        self.attach(ReshapeLayer(name='reshape'))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout1"))
        self.attach(InnerProductLayer(in_channel_num=256*5*5,
                                      out_channel_num=4096,
                                     init_para=init,
                                      wd=wd,
                                      name='ip1'))
        self.attach(ReLULayer(name='relu_ip1'))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout2"))
        self.attach(InnerProductLayer(in_channel_num=4096,
                                      out_channel_num=4096,
                                     init_para=init,
                                      wd=wd,
                                      name='ip2'))
        self.attach(ReLULayer(name='relu_ip2'))
        self.attach(InnerProductLayer(in_channel_num=4096,
                                      out_channel_num=self.in_channel_num,
                                     init_para=init,
                                      wd=wd,
                                      name='ip_last'))
        self.attach(SoftmaxWithLossLayer(
            class_num=self.in_channel_num,
            inputs=[{"name": "ip_last", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))


    def _build_cifar10_model(self):
        self.attach(ConvolutionLayer([5, 5],
                                     [1, 1, 1, 1],
                                     'SAME',
                                     init_para={
                                         "name": "truncated_normal",
                                         "stddev": 1e-4},
                                     wd={"type": "l2", "scale": 0},
                                     in_channel_num=3,
                                     out_channel_num=64,
                                     name='conv1'))
        self.attach(ReLULayer(name='relu1'))
        self.attach(PoolingLayer([1, 3, 3, 1],
                                 [1, 2, 2, 1],
                                 'SAME',
                                 name='pool1'))
        self.attach(LRNLayer(name='norm1'))

        self.attach(ConvolutionLayer([5, 5],
                                     [1, 1, 1, 1],
                                     'SAME',
                                     initial_bias_value=0.1,
                                     init_para={
                                         "name": "truncated_normal",
                                         "stddev": 1e-4},
                                     wd={"type": "l2", "scale": 0},
                                     in_channel_num=64,
                                     out_channel_num=64,
                                     name='conv2'))
        self.attach(ReLULayer(name='relu2'))
        self.attach(LRNLayer(name='norm2'))
        self.attach(PoolingLayer([1, 3, 3, 1],
                                 [1, 2, 2, 1],
                                 'SAME',
                                 name='pool2'))
        self.attach(InnerProductLayer(initial_bias_value=0.1,
                                      init_para={
                                          "name": "truncated_normal",
                                          "stddev": 0.04},
                                      wd={"type": "l2", "scale": 0.004},
                                      in_channel_num=self.in_channel_num,
                                      out_channel_num=384,
                                      name='ip1'))
        self.attach(ReLULayer(name='relu3'))

        self.attach(InnerProductLayer(initial_bias_value=0.1,
                                      init_para={
                                          "name": "truncated_normal",
                                          "stddev": 0.04},
                                      wd={"type": "l2", "scale": 0.004},
                                      in_channel_num=384,
                                      out_channel_num=192,
                                      name='ip2'))
        self.attach(InnerProductLayer(initial_bias_value=0,
                                      init_para={
                                          "name": "truncated_normal",
                                          "stddev": 1/192.0},
                                      wd={"type": "l2", "scale": 0},
                                      in_channel_num=192,
                                      out_channel_num=10,
                                      name='softmax_linear'))

        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[{"name": "softmax_linear", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))


class OneLayerBrain(GraphBrain):
    def __init__(self, **kwargs):
        super(OneLayerBrain, self).__init__(**kwargs)
        self.attach(
            ConvolutionLayer(ksize=[5, 5],
                             strides=[1, 1],
                             padding="SAME",
                             in_channel_num=1,
                             out_channel_num=32,
                             name="conv1")
        )
        self.attach(ReLULayer(name="relu1"))
        self.attach(
            PoolingLayer(ksize=[5, 5],
                         strides=[5, 5],
                         padding="SAME",
                         name="pool1")
        )

        self.attach(InnerProductLayer(in_channel_num=1152, out_channel_num=10, name="ip1"))
        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[
                {"name": "ip1", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
            name="loss"))


class LeNet(GraphBrain):
    """
    A rough LeNet. It is supposed to copy the example from Caffe, but for the
    time being it has not been checked whether they are exactly the same.
    """
    def __init__(self, **kwargs):
        super(LeNet, self).__init__(**kwargs)
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     in_channel_num=1,
                                     out_channel_num=32,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(PoolingLayer(ksize=[2, 2],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool1"))

        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     in_channel_num=32,
                                     out_channel_num=64,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(PoolingLayer(ksize=[5, 5],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool2"))

        self.attach(InnerProductLayer(
            in_channel_num=3136,
            out_channel_num=512,
            name="ip1"))
        self.attach(ReLULayer(name="relu3"))

        self.attach(InnerProductLayer(
            in_channel_num=512,
            out_channel_num=10,
            name="ip2"))

        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[{"name": "ip2", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))


class MnistTfTutorialNet(GraphBrain):
    """
    A multiple layer network with parameters from the MNIST tutorial of
    tensorflow.
    """
    def __init__(self, **kwargs):
        super(MnistTfTutorialNet, self).__init__(**kwargs)
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     initial_bias_value=0.,
                                     init_para={"name": "truncated_normal",
                                                "stddev": 0.1},
                                     wd={"type": "l2", "scale": 5e-4},
                                     in_channel_num=1,
                                     out_channel_num=32,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(PoolingLayer(ksize=[2, 2],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool1"))

        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     initial_bias_value=0.1,
                                     init_para={"name": "truncated_normal",
                                                "stddev": 0.1},
                                     wd={"type": "l2", "scale": 5e-4},
                                     in_channel_num=32,
                                     out_channel_num=64,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(PoolingLayer(ksize=[5, 5],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool2"))

        self.attach(InnerProductLayer(in_channel_num=3136,
                                      out_channel_num=512,
                                      initial_bias_value=0.1,
                                      init_para={"name": "truncated_normal",
                                                 "stddev": 0.1},
                                      wd={"type": "l2", "scale": 5e-4},
                                      name="ip1"))
        self.attach(ReLULayer(name="relu3"))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout1"))

        self.attach(InnerProductLayer(in_channel_num=512,
                                      out_channel_num=10,
                                      initial_bias_value=0.1,
                                      init_para={"name": "truncated_normal",
                                                 "stddev": 0.1},
                                      wd={"type": "l2", "scale": 5e-4},
                                      name="ip2"))

        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[{"name": "ip2", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))


class NewMnistTfTutorialNet(GraphBrain):
    """
    A multiple layer network with parameters from the MNIST tutorial of
    tensorflow. Changed the initialization methods from truncated normal to
    default.
    """
    def __init__(self, **kwargs):
        super(NewMnistTfTutorialNet, self).__init__(**kwargs)
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     initial_bias_value=0.,
                                     init_para={"name": "default"},
                                     wd={"type": "l2", "scale": 5e-4},
                                     in_channel_num=1,
                                     out_channel_num=32,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(PoolingLayer(ksize=[2, 2],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool1"))

        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     initial_bias_value=0.1,
                                     init_para={"name": "default"},
                                     wd={"type": "l2", "scale": 5e-4},
                                     in_channel_num=32,
                                     out_channel_num=64,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(PoolingLayer(ksize=[5, 5],
                                 strides=[2, 2],
                                 padding="SAME",
                                 name="pool2"))

        self.attach(InnerProductLayer(
            # The difference is caused by different way of doing padding in TF
            # and Torch.
            in_channel_num=3136 if A.backend() == A.TF else 2304,
            out_channel_num=512,
            initial_bias_value=0.1,
            init_para={"name": "default"},
            wd={"type": "l2", "scale": 5e-4},
            name="ip1"))
        self.attach(ReLULayer(name="relu3"))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout1"))

        self.attach(InnerProductLayer(in_channel_num=512,
                                      out_channel_num=10,
                                      initial_bias_value=0.1,
                                      init_para={"name": "default"},
                                      wd={"type": "l2", "scale": 5e-4},
                                      name="ip2"))

        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[{"name": "ip2", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))


class VGGNet(GraphBrain):
    def __init__(self,
                 class_num=10,
                 padding="SAME",
                 loss_layer=None,
                 **kwargs):
        """
        Args:
            loss_layer: A tuple of (A python Class, A dict).
                The type of loss layer to use in this net. The first item is
                the class of the loss layer, and the second is extra parameters
                of this layer, besides `name`. If None, a softmax cross_entropy
                loss will be used.
        """
        super(VGGNet, self).__init__(**kwargs)
        self.padding = padding

        # The number counted by how many convolution layer has been applied. It
        # is used to give a easily told name to each layer.
        self.top_layer_No = 0

        self.attach_conv_bn_relu(3, 64)
        self.attach(DropoutLayer(keep_prob=0.7,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(64, 64)
        self.attach(PoolingLayer(ksize=[1, 2, 2, 1],
                                 strides=[1, 2, 2, 1],
                                 padding=self.padding,
                                 name="pool{}".format(self.top_layer_No)))

        self.attach_conv_bn_relu(64, 128)
        self.attach(DropoutLayer(keep_prob=0.6,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(128, 128)
        self.attach(PoolingLayer(ksize=[1, 2, 2, 1],
                                 strides=[1, 2, 2, 1],
                                 padding=self.padding,
                                 name="pool{}".format(self.top_layer_No)))

        self.attach_conv_bn_relu(128, 256)
        self.attach(DropoutLayer(keep_prob=0.6,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(256, 256)
        self.attach(DropoutLayer(keep_prob=0.6,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(256, 256)
        self.attach(PoolingLayer(ksize=[1, 2, 2, 1],
                                 strides=[1, 2, 2, 1],
                                 padding=self.padding,
                                 name="pool{}".format(self.top_layer_No)))

        self.attach_conv_bn_relu(256, 512)
        self.attach(DropoutLayer(keep_prob=0.6,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(512, 512)
        self.attach(DropoutLayer(keep_prob=0.6,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach_conv_bn_relu(512, 512)
        self.attach(PoolingLayer(ksize=[1, 2, 2, 1],
                                 strides=[1, 2, 2, 1],
                                 padding=self.padding,
                                 name="pool{}".format(self.top_layer_No)))

        self.top_layer_No += 1
        self.attach(DropoutLayer(keep_prob=0.5,
                                 name="dropout{}".format(self.top_layer_No)))
        self.attach(InnerProductLayer(
            in_channel_num=2048,
            out_channel_num=512,
            init_para={
                "name": "truncated_normal",
                "stddev": 1e-4},
            name="ip1"))
        self.attach(
            BatchNormalizationLayer(
                channel_num=512,
                name="bn{}".format(self.top_layer_No)))
        self.attach(ReLULayer(name="relu{}".format(self.top_layer_No)))

        self.top_layer_No += 1
        self.attach(DropoutLayer(keep_prob=0.5,
                                 name="dropout{}".format(self.top_layer_No)))

        self.attach(InnerProductLayer(
            in_channel_num=512,
            out_channel_num=class_num,
            init_para={
                "name": "truncated_normal",
                "stddev": 1e-4},
            name="ip2"))
        self.attach(
            BatchNormalizationLayer(
                channel_num=class_num,
                name="bn{}".format(self.top_layer_No)))

        if loss_layer:
            self.attach(loss_layer[0](
                class_num=class_num,
                name="loss",
                **loss_layer[1]))
        else:
            self.attach(SoftmaxWithLossLayer(
                class_num=class_num,
                inputs=[{"name": "ip2", "idxs": [0]},
                        {"name": "system_in", "idxs": [1]}],
                name="loss"))

    def attach_conv_bn_relu(self, in_channel_num, out_channel_num):
        """
        This method attach a block of layer, aka convolution, batch
        normalization, and ReLU, to the brain. It also maintains
        `top_layer_No`.
        """
        self.top_layer_No += 1
        self.attach(ConvolutionLayer([3, 3],
                                     [1, 1, 1, 1],
                                     padding=self.padding,
                                     init_para={
                                         "name": "truncated_normal",
                                         "stddev": 1e-4},
                                     wd={"type": "l2", "scale": 5e-4},
                                     in_channel_num=in_channel_num,
                                     out_channel_num=out_channel_num,
                                     name="conv{}".format(self.top_layer_No)))
        self.attach(BatchNormalizationLayer(
            channel_num=out_channel_num,
            name="bn{}".format(self.top_layer_No)))
        self.attach(ReLULayer(name="relu{}".format(self.top_layer_No)))


class ResNet(GraphBrain):
    def __init__(self,
                 depth=28,
                 width=2,
                 class_num=10,
                 dropout_prob=None,
                 projection_shortcut=True,
                 use_gsmax=False,
                 group_size=4,
                 **kwargs):
        super(ResNet, self).__init__(**kwargs)

        self.depth = depth
        self.width = width
        self.class_num = class_num
        self.residual_block_No = 0
        self.dropout_prob = dropout_prob
        self.projection_shortcut = projection_shortcut
        self.use_bias = None
        self.use_gsmax = use_gsmax
        self.group_size = group_size

    def _attach_stack(self,
                      n_input_plane,
                      n_output_plane,
                      count,
                      stride,
                      act_before_residual,
                      block_type="basic"):
        if block_type == "basic":
            conv_params = [[(3, 3), stride, "SAME"],
                           [(3, 3), (1, 1), "SAME"]]
        elif block_type == "bottleneck":
            # The implementation of bottleneck layer is a little tricky, though
            # I think I make it clearer than Facebook's version that uses a
            # global variable. Besides that we choose the type of residual unit
            # using the string here, in the actual `_attach_block` method, the
            # `n_input_plane` and `n_output_plane` methods stick with their
            # essential meaning when not using a bottleneck unit. That's to say
            # the real channel number when 3X3 convolution is applied. The
            # enlarging of channel number happens internally, similar with how
            # it is done in the facebook version.
            conv_params = [[(1, 1), (1, 1), "SAME"],
                           [(3, 3), stride, "SAME"],
                           [(1, 1), (1, 1), "SAME"]]
        else:
            raise Exception("Block type {} is not supported.".format(block_type))

        self._attach_block(n_input_plane,
                           n_output_plane,
                           stride,
                           act_before_residual,
                           conv_params)
        for i in range(2, count+1):
            if block_type == "basic":
                conv_params = [[(3, 3), (1, 1), "SAME"],
                               [(3, 3), (1, 1), "SAME"]]
            elif block_type == "bottleneck":
                conv_params = [[(1, 1), (1, 1), "SAME"],
                               [(3, 3), (1, 1), "SAME"],
                               [(1, 1), (1, 1), "SAME"]]
            else:
                raise Exception("Block type {} is not supported.".format(block_type))

            self._attach_block(n_output_plane,
                               n_output_plane,
                               (1, 1),
                               False,
                               conv_params)

    def _attach_block(self,
                      n_input_plane,
                      n_output_plane,
                      stride,
                      act_before_residual,
                      conv_params):
        self.residual_block_No += 1

        main_branch_layer_name = self.blocks[-1].name

        is_bottleneck = False  # A flag for shortcut padding.
        for i, v in enumerate(conv_params):
            ksize, r_stride, padding = v

            in_channel_num = n_output_plane
            if i == 0:
                in_channel_num = n_input_plane

            self.attach(BatchNormalizationLayer(
                channel_num=in_channel_num,
                dim_num=2,
                name="bn_{}_{}".format(self.residual_block_No, i)))

            if self.use_gsmax and n_input_plane > 16:
                self.attach(GroupSoftmaxLayer(
                    group_size=self.group_size*n_input_plane/(16*self.width),
                    name="gsmax_{}_{}".format(self.residual_block_No, i)))
            else:
                self.attach(ReLULayer(name="relu_{}_{}".format(
                    self.residual_block_No, i)))

            if i == 0:
                if n_input_plane != n_output_plane and act_before_residual:
                    # At the first block of each BIG layer, two branches may
                    # share the same BN and activation function, thus bookkeep
                    # the branching layer name.
                    main_branch_layer_name = self.blocks[-1].name

            # We determine whether this is bottleneck layer by checking the
            # kernel size.
            if i == len(conv_params) - 1 and list(ksize) == [1, 1]:
                # This is the last layer of a bottleneck layer, we need to
                # increase the channel number back.
                is_bottleneck = True
                # In wide residual network, only bottleneck conv layer is
                # widened, thus the effect needs be offset back.
                out_channel_num = n_output_plane * 4 / self.width
            else:
                out_channel_num = n_output_plane

            if i != 0:
                if self.dropout_prob:
                    self.attach(DropoutLayer(
                        keep_prob=1-self.dropout_prob,
                        name="dropout_{}_{}".format(self.residual_block_No,
                                                    i)))
            self.attach(ConvolutionLayer(ksize,
                                         [1, r_stride[0], r_stride[1], 1],
                                         padding=padding,
                                         init_para={"name": "msra"},
                                         initial_bias_value=self.use_bias,
                                         wd=self.wd,
                                         in_channel_num=in_channel_num,
                                         out_channel_num=out_channel_num,
                                         name="conv_{}_{}".format(
                                             self.residual_block_No, i)))

        last_residual_layer_name = self.blocks[-1].name

        if n_input_plane != n_output_plane:
            if self.projection_shortcut:
                self.attach(ConvolutionLayer(
                    [1, 1],
                    [1, stride[0], stride[1], 1],
                    inputs=[{"name": main_branch_layer_name}],
                    padding="SAME",
                    init_para={"name": "msra"},
                    initial_bias_value=self.use_bias,
                    wd=self.wd,
                    in_channel_num=n_input_plane,
                    out_channel_num=out_channel_num,
                    name="conv_{}_shortcut".format(self.residual_block_No)))
            else:
                _ = (1, stride[0], stride[1], 1)
                if is_bottleneck:
                    in_channel_num = n_input_plane * 4
                else:
                    in_channel_num = n_input_plane

                self.attach(PoolingLayer(
                    ksize=_,
                    strides=_,
                    inputs=[{"name": main_branch_layer_name}],
                    padding="VALID",
                    type="avg",
                    name="pool_{}_shortcut".format(self.residual_block_No)))
                if A.DATA_FORMAT == "CHW":
                    padding = [0,(out_channel_num - in_channel_num) // 2, 0, 0]
                else:
                    padding = [0, 0, 0, (out_channel_num - in_channel_num) // 2]
                self.attach(PaddingLayer(
                    padding=padding,
                    name="pad_{}_shortcut".format(self.residual_block_No)))

            shortcut_layer_name = self.blocks[-1].name
        else:
            shortcut_layer_name = main_branch_layer_name

        self.attach(MergeLayer(inputs=[{"name": last_residual_layer_name},
                                       {"name": shortcut_layer_name}],
                               name="merge_{}".format(self.residual_block_No)))


class CifarResNet(ResNet):
    def __init__(self,
                 color_channel_num=3,
                 pool_size=8,
                 n_stages = None,
                 sub_class_multiplier_ratio=0.5,
                 h_loss=False,
                 **kwargs):
        super(CifarResNet, self).__init__(**kwargs)

        depth = self.depth

        assert((depth - 4) % 6 == 0), \
            "Depth must be 6 * n + 4."
        k = self.width
        if not n_stages:
            n_stages = [16, 16*k, 32*k, 64*k]
        assert (depth - 4) % 6 is 0
        n = (depth - 4) // 6
        if self.projection_shortcut:
            act_before_residual = [True, True, True]
            self.wd = {"type": "l2", "scale": 5e-4}
        else:
            act_before_residual = [True, False, False]
            self.wd = {"type": "l2", "scale": 0.0002}

        self.attach(ConvolutionLayer([3, 3],
                                     [1, 1, 1, 1],
                                     padding="SAME",
                                     init_para={"name": "msra"},
                                     wd=self.wd,
                                     in_channel_num=color_channel_num,
                                     out_channel_num=n_stages[0],
                                     initial_bias_value=self.use_bias,
                                     name="conv0"))

        self._attach_stack(n_input_plane=n_stages[0],
                           n_output_plane=n_stages[1],
                           count=n,
                           stride=(1, 1),
                           act_before_residual=act_before_residual[0])
        self._attach_stack(n_input_plane=n_stages[1],
                           n_output_plane=n_stages[2],
                           count=n,
                           stride=(2, 2),
                           act_before_residual=act_before_residual[1])
        self._attach_stack(n_input_plane=n_stages[2],
                           n_output_plane=n_stages[3],
                           count=n,
                           stride=(2, 2),
                           act_before_residual=act_before_residual[2])
        self.attach(BatchNormalizationLayer(n_stages[3], dim_num=2, name="bn_out"))
        if self.use_gsmax:
            self.attach(GroupSoftmaxLayer(
                group_size=self.group_size*(n_stages[3], n_stages[1]),
                name="gsmax_out"))
        else:
            self.attach(ReLULayer(name="relu_out"))
        self.attach(PoolingLayer(ksize=[1, pool_size, pool_size, 1],
                                 strides=[1, 1, 1, 1],
                                 padding="VALID",
                                 type="avg",
                                 name="global_pool"))
        self.attach(ReshapeLayer(name="reshape"))
        self.attach(InnerProductLayer(initial_bias_value=0,
                                      init_para={"name": "default"},
                                      wd=self.wd,
                                      in_channel_num=n_stages[3],
                                      out_channel_num=self.class_num,
                                      name='ip'))
        if h_loss:
            self.attach(SoftmaxWithLossLayer(
                class_num=self.class_num,
                multiplier=sub_class_multiplier_ratio,
                inputs=[{"name": "ip"},
                        {"name": "system_in", "idxs": [2]}],
                name="softmax"))
            self.attach(CollapseOutLayer(group_size=5,
                                         type="average_out",
                                         inputs=[
                                             {"name": "ip"}
                                         ],
                                         name="average_out"))
            self.attach(SoftmaxWithLossLayer(
                class_num=20,
                multiplier=1-sub_class_multiplier_ratio,
                inputs=[
                    {"name": "average_out"},
                    {"name": "system_in", "idxs": [1]}],
                name="super_class_loss"))
        else:
            self.attach(SoftmaxWithLossLayer(
                class_num=self.class_num,
                inputs=[{"name": "ip"},
                        {"name": "system_in", "idxs": [1]}],
                name="softmax"))


class ImagenetResNet(ResNet):
    def __init__(self, **kwargs):
        super(ImagenetResNet, self).__init__(**kwargs)

        k = self.width
        n_stages = [64, 64*k, 128*k, 256*k, 512*k]
        # Given imagenet needs a some non-regular conv-bn-relu-pool block at
        # the beginning of the branch, it is tailed by handle, and replaces the
        # first shared pre-activation block. The remaining shared
        # pre-activation block are still active.
        act_before_residual = [False, True, True, True]
        strides = [1, 2, 2, 2]
        # The configuration to train imagenet.
        cfg = {
         18  : [[2, 2, 2, 2], "basicblock"],
         34  : [[3, 4, 6, 3], "basicblock"],
         50  : [[3, 4, 6, 3], "bottleneck"],
         101 : [[3, 4, 23, 3], "bottleneck"],
         152 : [[3, 8, 36, 3], "bottleneck"],
         200 : [[3, 24, 36, 3], "bottleneck"],
        }
        n_depth, block_name = cfg[self.depth]
        self.wd = {"type": "l2", "scale": 1e-4}

        self.attach(ConvolutionLayer([7, 7],
                                     [1, 2, 2, 1],
                                     padding="SAME",
                                     init_para={"name": "msra"},
                                     wd=self.wd,
                                     out_channel_num=64,
                                     initial_bias_value=self.use_bias,
                                     name="conv0"))
        self.attach(BatchNormalizationLayer(dim_num=2, name="bn0"))
        self.attach(ReLULayer(name="relu0"))
        self.attach(PoolingLayer(ksize=[1, 3, 3, 1],
                                 strides=[1, 2, 2, 1],
                                 padding="SAME",
                                 type="max",
                                 name="pool0"))
        for i in range(4):
            self._attach_stack(n_input_plane=n_stages[i],
                               n_output_plane=n_stages[i+1],
                               count=n_depth[i],
                               stride=(strides[i], strides[i]),
                               act_before_residual=act_before_residual[i],
                               block_type="bottleneck")

        self.attach(ReLULayer(name="relu_final"))
        self.attach(PoolingLayer(ksize=[1, 7, 7, 1],
                                 strides=[1, 1, 1, 1],
                                 padding="VALID",
                                 type="avg",
                                 name="global_pool"))
        self.attach(ReshapeLayer(name="reshape"))
        self.attach(InnerProductLayer(initial_bias_value=0,
                                      init_para={"name": "default"},
                                      wd=self.wd,
                                      out_channel_num=self.class_num,
                                      name='ip'))
        self.attach(SoftmaxWithLossLayer(
            class_num=self.class_num,
            inputs=[{"name": "ip"},
                    {"name": "system_in", "idxs": [1]}],
            name="softmax"))


class PatriceNet(GraphBrain):
    """
    Reproduction of the paper: *Best Practices for Convolutional Neural
    Networks Applied to Visual Document Analysis*.
    """
    def __init__(self, **kwargs):
        super(PatriceNet, self).__init__(**kwargs)

        self.attach(InnerProductLayer(in_channel_num=784,
                                      out_channel_num=800,
                                      # init_para={"name": "normal",
                                      #            "stddev": 0.05},
                                      initial_bias_value=0,
                                      wd=None,
                                      name="ip1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(InnerProductLayer(in_channel_num=800,
                                      out_channel_num=10,
                                      initial_bias_value=0,
                                      # init_para={"name": "normal",
                                      #            "stddev": 0.05},
                                      wd=None,
                                      name="ip2"))
        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[
                {"name": "ip2", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
            name="loss"))


class IntriguingConvNet(GraphBrain):
    """
    The modified version of simple convnet on MNIST from *Intriguing properties
    of neural networks*.
    """
    def __init__(self, **kwargs):
        super(IntriguingConvNet, self).__init__(**kwargs)

        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     in_channel_num=1,
                                     out_channel_num=32,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[2, 2],
                                     padding="SAME",
                                     in_channel_num=32,
                                     out_channel_num=64,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(InnerProductLayer(in_channel_num=10816,
                                      out_channel_num=64,
                                      initial_bias_value=0,
                                      # init_para={"name": "normal",
                                      #            "stddev": 0.05},
                                      # wd=None,
                                      name="ip1"))
        self.attach(ReLULayer(name="relu3",
                              summarize_output=True,
        ))
        self.attach(InnerProductLayer(in_channel_num=64,
                                      out_channel_num=10,
                                      initial_bias_value=0,
                                      # init_para={"name": "normal",
                                      #            "stddev": 0.05},
                                      wd=None,
                                      name="ip2"))
        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[
                {"name": "ip2", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
            name="loss"))


class AllConvNet(GraphBrain):
    """
    All Convolutional Network on CIFAR-10.
    """
    def __init__(self, **kwargs):
        super(AllConvNet, self).__init__(**kwargs)

        self.attach(DropoutLayer(keep_prob=0.8, name="dropout1"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=3,
                                     out_channel_num=96,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=96,
                                     out_channel_num=96,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[2, 2],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=96,
                                     out_channel_num=96,
                                     name="conv3"))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout2"))
        self.attach(ReLULayer(name="relu3"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=96,
                                     out_channel_num=192,
                                     name="conv4"))
        self.attach(ReLULayer(name="relu4"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=192,
                                     out_channel_num=192,
                                     name="conv5"))
        self.attach(ReLULayer(name="relu5"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[2, 2],
                                     padding="SAME",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=192,
                                     out_channel_num=192,
                                     name="conv6"))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout3"))
        self.attach(ReLULayer(name="relu6"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="VALID",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=192,
                                     out_channel_num=192,
                                     name="conv7"))
        self.attach(ReLULayer(name="relu7"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="VALID",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=192,
                                     out_channel_num=192,
                                     name="conv8"))
        self.attach(ReLULayer(name="relu8"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="VALID",
                                     init_para={"name": "xavier"},
                                     wd={"type": "l2", "scale": 0.001},
                                     in_channel_num=192,
                                     out_channel_num=10,
                                     name="conv9"))
        self.attach(ReLULayer(name="relu9"))
        self.attach(PoolingLayer(ksize=[6, 6],
                                 strides=[1, 1],
                                 padding="VALID",
                                 type="avg",
                                 name="global_pool"))
        self.attach(ReshapeLayer(name="reshape"))
        self.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[
                {"name": "reshape", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
            name="loss"))

class AllConvImagenet(GraphBrain):
    """
    All convolutional style AlexNet on ImageNet.
    """
    def __init__(self, **kwargs):
        super(AllConvImagenet, self).__init__(**kwargs)
        wd = {"type": "l2", "scale": 1 * 10e-4}
        # wd = None
        # init = {"name": "msra"}
        init = {"name": "xavier"}
        self.attach(ConvolutionLayer(ksize=[11, 11],
                                     strides=[4, 4],
                                     padding="VALID",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=3,
                                     out_channel_num=96,
                                     name="conv1"))
        self.attach(ReLULayer(name="relu1"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="VALID",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=96,
                                     out_channel_num=96,
                                     name="conv2"))
        self.attach(ReLULayer(name="relu2"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[2, 2],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=96,
                                     out_channel_num=96,
                                     name="conv3"))
        self.attach(ReLULayer(name="relu3"))
        self.attach(ConvolutionLayer(ksize=[5, 5],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=96,
                                     out_channel_num=256,
                                     name="conv4"))
        self.attach(ReLULayer(name="relu4"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=256,
                                     out_channel_num=256,
                                     name="conv5"))
        self.attach(ReLULayer(name="relu5"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[2, 2],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=256,
                                     out_channel_num=256,
                                     name="conv6"))
        self.attach(ReLULayer(name="relu6"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=256,
                                     out_channel_num=384,
                                     name="conv7"))
        self.attach(ReLULayer(name="relu7"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=384,
                                     out_channel_num=384,
                                     name="conv8"))
        self.attach(ReLULayer(name="relu8"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[2, 2],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=384,
                                     out_channel_num=384,
                                     name="conv9"))
        self.attach(DropoutLayer(keep_prob=0.5, name="dropout1"))
        self.attach(ReLULayer(name="relu9"))
        self.attach(ConvolutionLayer(ksize=[3, 3],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=384,
                                     out_channel_num=1024,
                                     name="conv10"))
        self.attach(ReLULayer(name="relu10"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=1024,
                                     out_channel_num=1024,
                                     name="conv11"))
        self.attach(ReLULayer(name="relu11"))
        self.attach(ConvolutionLayer(ksize=[1, 1],
                                     strides=[1, 1],
                                     padding="SAME",
                                     init_para=init,
                                     wd=wd,
                                     in_channel_num=1024,
                                     out_channel_num=1000,
                                     name="conv12"))
        self.attach(ReLULayer(name="relu12"))
        self.attach(PoolingLayer(ksize=[5, 5],
                                 strides=[1, 1],
                                 padding="VALID",
                                 type="avg",
                                 name="global_pool"))
        self.attach(ReshapeLayer(name="reshape"))
        self.attach(SoftmaxWithLossLayer(
            class_num=1000,
            inputs=[
                {"name": "reshape", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
            name="loss"))



"""
This module contains templates to build network blocks that consists of several
usual combination of layer types. For instance, a block of convolutional neural
network normally consists of a convolution layer, a pooling layer and an
activation layer. Those templates server as syntax sugars to
ease writing network structures. So instead of appending three layers and
organizing their in a statement block, one could just use a template to create
a block.
"""
from __future__ import absolute_import
from __future__ import print_function
import sys
import inspect

from akid.layers import (
    ConvolutionLayer,
    PoolingLayer,
    InnerProductLayer,
    DropoutLayer,
    CollapseOutLayer,
    ReLULayer,
    SoftmaxNormalizationLayer,
    BatchNormalizationLayer,
    GroupSoftmaxLayer
)


# A counter to give number count to layer names
counter = 0


def init():
    """
    Call this to reset counter value if more than one networks are being built
    in a running instance.
    """
    global counter
    counter = 0


def cnn_block(ksize=None,
              conv_padding="SAME",
              initial_bias_value=0.,
              init_para={"name": "truncated_normal",
                         "stddev": 0.1},
              wd={"type": "l2", "scale": 5e-4},
              max_norm=None,
              in_channel_num=32,
              out_channel_num=32,
              pool_size=[2, 2],
              pool_stride=[2, 2],
              activation={"type": "relu"},
              activation_before_pooling=False,
              keep_prob=None,
              bn=None):
    global counter
    counter += 1
    block = []

    if ksize:
        block.append(ConvolutionLayer(ksize=ksize,
                                      strides=[1, 1],
                                      padding=conv_padding,
                                      initial_bias_value=initial_bias_value,
                                      init_para=init_para,
                                      wd=wd,
                                      max_norm=max_norm,
                                      in_channel_num=in_channel_num,
                                      out_channel_num=out_channel_num,
                                      name="conv{}".format(counter)))
        if pool_size:
            block.append(PoolingLayer(ksize=pool_size,
                                      strides=pool_stride,
                                      padding="SAME",
                                      name="pool{}".format(counter)))
    else:
        block.append(InnerProductLayer(out_channel_num=out_channel_num,
                                       in_channel_num=in_channel_num,
                                       initial_bias_value=initial_bias_value,
                                       init_para=init_para,
                                       wd=wd,
                                       name="ip{}".format(counter)))
    if bn:
        bn_layer = BatchNormalizationLayer(
            name="bn{}".format(counter), **bn)

    if activation:
        try:
            activation_type = activation["type"]
            if activation_type == "relu":
                layer = ReLULayer(name="relu{}".format(counter))
            elif activation_type == "maxout":
                layer = CollapseOutLayer(name="maxout{}".format(counter),
                                         group_size=activation["group_size"])
            elif activation_type == "ngsmax":
                layer = SoftmaxNormalizationLayer(
                    name="ngsoftmax{}".format(counter),
                    group_size=activation["group_size"])
            elif activation_type == "gsmax":
                layer = GroupSoftmaxLayer(name="gsmax{}".format(counter),
                                          group_size=activation["group_size"])
            else:
                print(("{} activation type has not been supported"
                       " yet.".format(activation_type)))
                sys.exit(0)
        except KeyError as e:
                print(("{} not found. You perhaps have a typo or miss a"
                       " parameter.".format(e)))
                sys.exit(0)
    else:
        layer = None

    if activation_before_pooling:
        for i, b in enumerate(block):
            if type(b) is PoolingLayer:
                if layer:
                    block.insert(i, layer)
                if bn:
                    block.insert(i, bn_layer)
                break
    else:
        # Even if there is no activation layer, which happens at the last
        # readout layer, BN may still be needed.
        if bn:
            block.append(bn_layer)
        if layer:
            block.append(layer)

    if keep_prob:
        block.append(DropoutLayer(keep_prob=keep_prob,
                                  name="dropout{}".format(counter)))

    return block

__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



from __future__ import absolute_import
import tensorflow as tf

from akid import AKID_DATA_PATH
from akid.core import kids, kongfus
from akid import ImagenetTFSource
from akid import IntegratedSensor
from akid.models.brains import ImagenetResNet

brain = ImagenetResNet(depth=50, width=2, class_num=1001,
                       name='WRN-50-2')

# Set up a sensor.
# #########################################################################
source = ImagenetTFSource(
    has_super_label=False,
    name="Imagenet",
    url=None,
    work_dir=AKID_DATA_PATH + "/imagenet",
    num_train=1281167,
    num_val=50000)

from akid.core.jokers import (
    Joker,
    CropJoker,
    FlipJoker,
    ResizeJoker
)

sensor = IntegratedSensor(source_in=source,
                          min_fraction_of_examples_in_queue=0.01,
                          num_preprocess_threads=4 * 8,
                          batch_size=256,
                          val_batch_size=200,
                          name='data')
# resize joker
class ColorJoker(Joker):
    def _forward(self, data_in):
      data = tf.image.random_brightness(data_in, max_delta=32. / 255.)
      data = tf.image.random_saturation(data, lower=0.5, upper=1.5)
      data = tf.image.random_hue(data, max_delta=0.2)
      data = tf.image.random_contrast(data, lower=0.5, upper=1.5)
      data = tf.clip_by_value(data, 0.0, 1.0)
      self._data = data


sensor.attach(ResizeJoker(224, 224, name="resize"))
sensor.attach(FlipJoker(flip_left_right=True, name="left_right_flip"))
sensor.attach(ColorJoker(name="color_distortion"))

sensor.attach(
    CropJoker(center=True, central_fraction=0.875, name="center_crop"),
    to_val=True)
sensor.attach(ResizeJoker(224, 224, name="resize"), to_val=True)



class AdhocJoker(Joker):
    def _forward(self, data_in):
        # Make the range of image be [-1, 1]
        data = tf.subtract(data_in, 0.5)
        data = tf.multiply(data, 2.0)
        data = tf.cast(data, tf.float32)
        # because of jpeg decoding, the shape is determined dynamically, thus
        # we manually set shape here.
        data = tf.reshape(data, shape=[224, 224, 3])
        self._data = data


sensor.attach(AdhocJoker(name="to_1_-1"))
sensor.attach(AdhocJoker(name="to_1_-1"), to_val=True)

def update_lr(kid):
    if kid.epoch < 30:
        kid.kongfu.lr_value = 0.1
    elif kid.epoch < 60:
        kid.kongfu.lr_value = 0.01
    else:
        kid.kongfu.lr_value = 0.001
# Summon a survivor.
# #########################################################################
from akid import LearningRateScheme
kid = kids.Kid(
    sensor,
    brain,
    kongfus.MomentumKongFu(lr_scheme={"name": LearningRateScheme.placeholder}),
    engine={"name": "data_parallel", "num_gpu": 8},
    # engine={"name": "single"},
    log_dir="log",
    max_epoch=100)
kid.hooks.on_batch_begin.append(update_lr)
kid.setup()
kid.practice()



from __future__ import absolute_import
import torch as th


def compute_gradients(opt, v):
    opt.zero_grad()
    v.backward()


def apply_gradients(opt, grads):
    return opt.step()


def MomentumOptimizer(lr, var_list=None, momentum=0.9, use_nesterov=False):
    return th.optim.SGD(var_list, lr, momentum, nesterov=use_nesterov)


def GradientDescentOptimizer(lr, var_list=None, momentum=0, use_nesterov=False):
    return th.optim.SGD(var_list, lr, momentum, nesterov=use_nesterov)



"""
The interface between nature and an (artificial) signal processing system,
saying a brain, is a `Sensor`. It does data augmentation (if needed), and
batches datum from `Source`.

Strictly speaking, the functional role of a sensor is to convert the signal in
the natural form to a form the data processing engine, which is the brain in
this case, could process. It is a Analog/Digital converter. However, the input
from `Source` is already in digital form, so this function is not there
anymore. But the data batching, augmentation and so on could still be put in
preprocessing. Thus we still use the name sensor for concept reuse.

Sensors can also include the support for functions such as multi-threaded, or
multi-process data prefetch to build better performed data preparation
pipeline.

Mathematically, it is a system made up with a series of linked blocks that do
data augmentation.

As an example, again saying in supervised setting, a sensor is a block that
takes a data source and output sensed (batched and augmented) data. A sensor
needs to be used along with a source. A concrete example could be the sensor
for the MNIST dataset. Taking a `Source`, we could make a sensor::

      sensor = SimpleSensor(name='data', source_in=source)

Similar with `Source`, a sensor has multiple modes to supply data. Refer to the
documentation of `Source` for more details.

------------------
Sensor as iterator
------------------

A `Sensor` can be used as an iterator when an automatic precise control on many
epochs of data should be provided is needed. To use sensor as an iterator::

    for batch in sensor:
        # blah blah

Note that, even if a `Sensor` is used as an iterator, one should call `setup`
and `teardown` before a sensor is used and after it finishes its role. This is
to properly shutdown threads, or processes that are prefetching data, if there
are any.

NOTE: Some sensors may do not support iteration. Check documentation for details.
"""
from __future__ import absolute_import, division, print_function

import sys
import abc
import os
import inspect
from deprecated import deprecated
import six
from six.moves import range

if sys.version_info[0] == 2:
    from six.moves.queue import Queue
    import six.moves.queue as queue
else:
    from queue import Queue
    import queue
import threading

import tensorflow as tf
import torch as th
from torch import multiprocessing as mp

from .jokers import JokerSystem
from .blocks import ValidatableProcessingBlock
from . import sources
from . import samplers
from .common import TRAIN_SUMMARY_COLLECTION, VALID_SUMMARY_COLLECTION
from .events import (
    DataPrefetchThreadsDeadEvent,
    DataPrefetchProcessesDeadEvent,
    EpochCompletedEvent
)
from akid import backend as A
from akid.utils import glog as log


def _check_if_main_thread_is_dead():
    # If the main thread is dead, quit.
    for t in threading.enumerate():
        if t.name == "MainThread":
            if not t.is_alive():
                return True
            else:
                return False


def _data_fetching_worker(source, index_queue, data_queue, done_event):
    # If index queue is not empty, then fetch the indices of a batch, and load
    # the data to data queue.
    while True:
        if _check_if_main_thread_is_dead():
            return

        # If the done event is set, quit.
        if done_event.is_set():
            return

        try:
            log.debug("Data Prefetching Worker: Fetching data according to indices ... ")
            indices = index_queue.get(timeout=A.TIMEOUT)
            data = source.get(indices)
            data_gpu = [A.Tensor(d) for d in data]
            data_queue.put(data_gpu, timeout=A.TIMEOUT)
            log.debug("Data Prefetching Worker: Data fetched")
        except queue.Empty:
            continue


def _data_fetching_worker_process(i, source, index_queue, data_queue, done_event):
    repeat = False
    NAME = "Data Prefetching Worker {}".format(i)

    while True:
        log.debug("{}: Working ...".format(NAME))
        try:
            os.kill(os.getppid(), 0)
        except ProcessLookupError:
            return

        # If the done event is set, quit.
        log.debug("{}: Done event {}".format(NAME, done_event.is_set()))
        if done_event.is_set():
            return

        try:
            log.debug("{}: Fetching data according to indices ... ".format(NAME))
            if not repeat:
                # It is possible the previous try to enqueue data is timeout,
                # so we do not fetch data this time.
                item = index_queue.get(timeout=A.TIMEOUT)
                if type(item) is EpochCompletedEvent:
                    # Epoch finishes, we pass on the event to let the sensor
                    # knows an epoch has indeed finished.
                    data = item
                else:
                    data = source.get(item) # The item is a list of indices now.
            data_queue.put(data, timeout=A.TIMEOUT)
            repeat = False
            log.debug("{}: Data fetched".format(NAME))
        except queue.Empty:
            continue
        except queue.Full:
            repeat = True
            continue
        except Exception as e:
            log.debug("Process {} : Exception {}.".format(i, e))
            raise e


def _data_preloading_worker(prefetch_data_queue, data_queue, done_event):
    log.debug("Data Preloading Worker: Started. ")
    NAME = "Data Preloading Worker"

    repeat = False

    while True:
        log.debug("Data Preloading Worker: Working ... ")
        if _check_if_main_thread_is_dead():
            return

        try:
            log.debug("{}: Loading CPU data ... ".format(NAME))

            if not repeat or done_event.is_set():
                data = prefetch_data_queue.get(timeout=A.TIMEOUT)
                if data is None:
                    assert done_event.is_set()
                    return
                elif done_event.is_set():
                    # We need to consume all data before finishing, otherwise,
                    # data fetching workers may cannot finish since it waits to
                    # enqueue data (for that the pipe used by the queue may be
                    # full, thus the workers are waiting the data to be
                    # consumed; this is an implementation issue of
                    # multiprocessing.Queue in python). But we do not need to
                    # process them anymore.
                    continue
                elif type(data) is list:
                    data_gpu = []
                    for d in data:
                        data_gpu.append(A.Tensor(d))

            log.debug("{}: Enqueuing GPU data ... ".format(NAME))

            if type(data) is EpochCompletedEvent:
                data_queue.put(data, timeout=A.TIMEOUT)
            else:
                data_queue.put(data_gpu, timeout=A.TIMEOUT)
            repeat = False

            log.debug("Data Preloading Worker: Data loaded")
        except queue.Empty:
            log.debug("{}: Queue is empty. Repeat.".format(NAME))
            continue
        except queue.Full:
            repeat = True
            continue


class Sensor(ValidatableProcessingBlock):
    """
    The top level abstract sensor to prepare data.

    The main thread would put indices to prefetch in an index queue. A data
    queue should be implemented by any subclass to store the data prefetched.

    By the default, a sensor is in the "train" mode. To use it in other modes,
    one needs to change the mode first, then set the batch size in that mode.
    """
    NAME = "Sensor"

    def __init__(self,
                 source_in=None,
                 batch_size=100,
                 val_batch_size=100,
                 test_batch_size=None,
                 queue_size=5,
                 sampler="shuffle",
                 val_sampler="sequence",
                 **kwargs):
        """
        Args:
            source_in: Source
                Where data should be sensed from.
            batch_size: int
                The number of samples in a time the sensor would provide.
            queue_size: int
                The number of batches to prefetch.
            sampler: str or a Sampler object
                The sampler that determines the sequence to process data.
            name: str
                Name of this sensor.
        """
        super(Sensor, self).__init__(**kwargs)
        self.source = source_in
        self.batch_size_dict = { A.Mode.TRAIN: batch_size,
                                 A.Mode.VAL: val_batch_size,
                                 A.Mode.TEST: val_batch_size if test_batch_size is None else test_batch_size}
        self.queue_size = queue_size
        if self.queue_size > self.num_batches_per_epoch:
            self.queue_size = self.num_batches_per_epoch

        # More complex datasets might customized samplers, Sensor supports pass
        # a built sampler directly.
        if type(sampler) is str:
            self.train_sampler_name = sampler
        else:
            self.train_sampler = sampler

        if type(val_sampler) is str:
            self.val_sampler_name = val_sampler
        else:
            self.val_sampler = val_sampler

        self.mode = A.Mode.TRAIN

    @property
    def batch_size(self):
        return self.batch_size_dict[self.mode]

    @property
    def num_batches_per_epoch(self):
        return (self.source.size - 1) // self.batch_size_dict[self.mode] + 1

    def set_batch_size(self, size):
        """
        NOTE: the queue size is determined dynamically from batch
        size. However, if batch size is set after sensor setup, the queue size
        is determined actually using the old batch size value.
        """
        if type(size) is not int:
            raise ValueError("The batch size should be of type int. Type {} received".format(type(size)))
        else:
            self.batch_size_dict[self.mode] = size
            if self.queue_size > self.num_batches_per_epoch:
                self.queue_size = self.num_batches_per_epoch

    def set_mode(self, mode):
        A.check_mode(mode)

        # We need to tear down pre-fetching threads before changing the mode,
        # otherwise, source would change mode queue shutdown, thus leading to
        # the result that some new data are fetched to the old data queue, and
        # not used.
        if self.is_setup:
            self._teardown_data_queue()

        self.mode = mode
        self.source.set_mode(mode)
        self.log("Mode {}".format(mode))

    def reset(self):
        if self.is_setup:
            self._teardown_data_queue()
            self.sampler.reset()

        self.setup()

    def _setup(self):
        """
        Set up a FIFO queue to load data from source.
        """
        self.source.setup()
        if self.mode == A.Mode.TRAIN:
            if hasattr(self, "train_sampler"):
                self.sampler = self.train_sampler
            else:
                self.sampler = samplers.get(self.train_sampler_name, self.source.size)
        elif self.mode == A.Mode.VAL or self.mode == A.Mode.TEST:
            if hasattr(self, "val_sampler"):
                self.sampler = self.val_sampler
            else:
                self.sampler = samplers.get(self.val_sampler_name, self.source.size)
        else:
            raise ValueError("Wrong mode {}".format(self.mode))

        self._setup_index_queue()
        self._setup_data_queue()

    def teardown(self):
        self._teardown_data_queue()

    @abc.abstractmethod
    def _setup_index_queue(self):
        """
        The procedure to set up an index queue.
        """
        pass

    @abc.abstractproperty
    def index_queue(self):
        """
        Any subclass should has a queue to provide indices for workers to
        fetch.
        """
        pass

    @property
    def data(self):
        return self._data

    @abc.abstractmethod
    def _setup_data_queue(self):
        """
        The procedure to set up a data queue should be implemented here.
        """
        pass

    @abc.abstractmethod
    def _teardown_data_queue(self):
        """
        The procedure to release resources allocated related to the data queue.
        """
        pass

    @abc.abstractproperty
    def data_queue(self):
        """
        Any subclass should has a data queue to hold data fetched according to
        index queue.
        """
        pass

    def _forward(self, *args, **kwargs):
        if self.index_queue.empty():
            try:
                # If we just finishes using the sensor as an iterator, calling
                # forward now would results in errors, since no data is in the data
                # queue now. If this is the case, we need to put same indices to
                # prefetch to continue.
                self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
            except EpochCompletedEvent:
                # In some cases, it is possible that an epoch is finished, yet
                # the next epoch has not been started. In such cases, we just
                # keep fetching.
                self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))

        ret = self.data_queue.get()

        A.cache_tensor_auto_scope(ret[0], "val_data" if self.is_val else "data")
        A.cache_tensor_auto_scope(ret[1], "val_labels" if self.is_val else "labels")
        self._data = ret

        # It looks like the first index queue refill could be merged with the
        # second one below. But it would require the index queue size is not
        # fill before each data queue fetch. Though this is an easy condition
        # to meet, I would like to just keep it in the current way.
        try:
            self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
        except EpochCompletedEvent:
            # Just keep prefetching the next batch.
            self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))

        return self._data

    def _post_forward(self, *args, **kwargs):
        super(Sensor, self)._post_forward(*args, **kwargs)

        if self.done_first_pass and not self.is_val:
            return

        if not self.do_summary:
            return

        if not self.is_val:
            self._image_summary(A.get_name(self.data[0]),
                                self.data,
                                TRAIN_SUMMARY_COLLECTION)
            return

        if self.is_val and not self.done_first_pass_val and self.do_summary_on_val:
            self._image_summary(A.get_name(self.data[0]),
                                self.data,
                                VALID_SUMMARY_COLLECTION)

    def _image_summary(self, name, image_batch, collection):
            self.log("Do tensorboard image summary on outputs {} of {}".format(
                name, self.name))

            A.summary.histogram(name,
                                image_batch,
                                collections=[collection])
            A.summary.image(name,
                            image_batch,
                            collections=[collection])


class SimpleSensor(Sensor):
    """
    A simple sensor that uses a single thread to prefetch data.
    """
    def _pre_forward(self):
        super(SimpleSensor, self)._pre_forward()
        if threading.active_count() == 1:
            # Only the main thread is alive. Raise exception.
            raise DataPrefetchThreadsDeadEvent

    def _setup_index_queue(self):
        # Set up a queue.
        self._index_queue = Queue(self.queue_size)
        # Start loading data from source according to mode.
        ## Put indices to prefetch.
        for i in range(self.queue_size):
            try:
                self._index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
            except EpochCompletedEvent:
                self._index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))

    @property
    def index_queue(self):
        return self._index_queue

    def _setup_data_queue(self):
        # Set up a data queue.
        self._data_queue = Queue(self.queue_size)
        # Start loading data from source according to mode.
        ## Start the workers to fetch data.
        self.done_event = threading.Event()
        self.worker_thread = threading.Thread(
            target=_data_fetching_worker,
            args=(self.source, self._index_queue, self._data_queue, self.done_event))
        self.worker_thread.start()

    def _teardown_data_queue(self):
        self.done_event.set()
        self.worker_thread.join()

    @property
    def data_queue(self):
        return self._data_queue


class ParallelSensor(Sensor):
    """
    The sensor that fetches data using multiple processes. It supports to be
    used as an iterator.
    """
    def __init__(self, num_workers=4, *args, **kwargs):
        super(ParallelSensor, self).__init__(*args, **kwargs)
        self.num_workers = num_workers
        self.queue_size *= num_workers
        if self.queue_size > self.num_batches_per_epoch:
            self.queue_size = self.num_batches_per_epoch

    def __iter__(self):
        self.epoch_finished = False
        if self.index_queue.empty():
            try:
                # Upon setup of sensor, we have some indices put in the queue, so
                # it may not be necessary put some indices. However, after the
                # second entry to create an iterator, it is possible the index
                # queue is empty, since further prefetching is disabled. In such a
                # case, we need to put some indices to prefetch to get started.
                self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
            except EpochCompletedEvent as e:
                # If the dataset is small, we could finish the dataset while we
                # are creating the iterator (since we are prefetching data upon
                # setup). In such a case, we would call it a day.
                self.epoch_finished = True
                self.index_queue.put(e)

        return self

    def __next__(self):
        # Get the next batch of data.
        ret = self.data_queue.get(timeout=A.TIMEOUT)

        # Raise `StopIteration` since we have received the epoch completion
        # signal we sent earlier.
        if type(ret) is EpochCompletedEvent:
            raise StopIteration

        # If not, we do our normal data stuff.
        A.cache_tensor_auto_scope(ret[0], "val_data" if self.is_val else "data")
        A.cache_tensor_auto_scope(ret[1], "val_labels" if self.is_val else "labels")
        self._data = ret

        # Put the indices of the batch to be prefetched if we still have not
        # finished an epoch. But if we have, stop fetching data by putting the
        # EpochCompletedEvent to the index queue.
        if not self.epoch_finished:
            try:
                self.index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
            except EpochCompletedEvent as e:
                self.epoch_finished = True
                self.index_queue.put(e)

        return ret

    def _pre_forward(self):
        super(ParallelSensor, self)._pre_forward()
        if len(mp.active_children()) == 0:
            # If all child processes are dead, we raise error.
            raise DataPrefetchProcessesDeadEvent
        if not self.preloading_thread.is_alive():
            # The thread to move data from CPU to GPU is dead.
            raise DataPrefetchThreadsDeadEvent

    def _setup_index_queue(self):
        self._index_queue = mp.Queue(self.queue_size)
        for i in range(self.queue_size):
            try:
                self._index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))
            except EpochCompletedEvent:
                self._index_queue.put(self.sampler.next(self.batch_size_dict[self.mode]))

    @property
    def index_queue(self):
        return self._index_queue

    def _setup_data_queue(self):
        # Set up a data queue.
        self._prefetch_data_queue = mp.Queue(self.queue_size)
        # Start loading data from source according to mode.
        ## Start the workers to fetch data.
        self.done_event = mp.Event()
        self.worker_processes = []
        for i in range(self.num_workers):
            process = mp.Process(
                target=_data_fetching_worker_process,
                args=(i, self.source, self._index_queue, self._prefetch_data_queue, self.done_event))
            process.start()
            self.worker_processes.append(process)

        # Start a thread to load the CPU data to GPU.
        self._data_queue = Queue(self.queue_size * 2)
        self.preloading_thread = threading.Thread(
            target=_data_preloading_worker,
            args=(self._prefetch_data_queue, self._data_queue, self.done_event))
        self.preloading_thread.start()

    def _teardown_data_queue(self):
        self.done_event.set()
        for i in range(self.num_workers):
            self.log("Waiting worker {} to join ...".format(i))
            self.worker_processes[i].join()
        # When training on MNIST, the preloading_thread dies each time data
        # queues are torn down. To prevent deadlock, check whether it is dead
        # or not. Why it dies is not clear.
        if self.preloading_thread.is_alive():
            self._prefetch_data_queue.put(None)
            self.log("Waiting preloading thread to join ...")
            self.preloading_thread.join()
        else:
            self.log("Preloading thread is dead already.")

    @property
    def data_queue(self):
        return self._data_queue


@deprecated(reason="Legacy code. Use `Sensor` instead.")
class OldSensor(six.with_metaclass(abc.ABCMeta, ValidatableProcessingBlock)):
    """
    The top level abstract sensor to preprocessing raw data received from
    `Source`, such as batching, data augmentation etc.
    """

    def __init__(self,
                 source_in=None,
                 batch_size=100,
                 val_batch_size=100,
                 shuffle_train=True,
                 **kwargs):
        """
        Args:
            name: str
                Name of this sensor.
            source_in: Source
                Where data should be sensed from.
            batch_size: int
                The number of samples a time the sensor would provide.
            val_batch_size: int
                The number of samples a time the sensor would provide when
                doing validation. It is supposed to evenly divide the number of
                validation samples.
        """
        super(OldSensor, self).__init__(self, **kwargs)
        self.batch_size = batch_size
        self.val_batch_size = val_batch_size
        self.source = source_in
        self.shuffle_train = shuffle_train
        self.done_first_pass_val = False

        if issubclass(type(self.source), sources.StaticSource):
            self.num_batches_per_epoch_train \
                = (self.source.num_train - 1) // self.batch_size + 1
            self.num_batches_per_epoch_val \
                = (self.source.num_val - 1) // self.val_batch_size + 1
            self.log("A epoch of training set contains {} batches".format(
                self.num_batches_per_epoch_train))
            self.log("A epoch of validation set contains {} batches".format(
                self.num_batches_per_epoch_val))

    @property
    def data(self):
        if self.is_val:
            return self.val_data
        else:
            return self.training_data

    @property
    def labels(self):
        if issubclass(type(self.source), sources.SupervisedSource):
            if self.is_val:
                return self.val_labels
            else:
                return self.training_labels
        else:
            raise Exception("OldSensor {} is not supposed to provide"
                            "labels".format(self.name))

    @abc.abstractmethod
    def _forward_train(self):
        """
        An abstract method each subclass must implement to provide training
        data.

        Returns:
            A tuple of tensors, (training_data, training_labels), should be
            returned if the source of this sensor is a supervised one;
            otherwise, a tensor of training data should be returned.
        """
        raise NotImplementedError("Each sensor needs to implement the method"
                                  " to actually provide data as tensors.")
        sys.exit()

    @abc.abstractmethod
    def _forward_val(self):
        """
        An abstract method each subclass must implement to provide validation
        data.

        Returns:
            A tuple of tensors, (training_data, training_labels), should be
            returned if the source of this sensor is a supervised one;
            otherwise, a tensor of training data should be returned.
        """
        raise NotImplementedError("Each sensor needs to implement the method"
                                  " to actually provide data as tensors.")

    def _post_forward(self, *args, **kwargs):
        super(OldSensor, self)._post_forward(*args, **kwargs)

        if self.done_first_pass and not self.is_val:
            return

        if not self.do_summary:
            return

        if not self.is_val\
           or (self.is_val\
               and not self.done_first_pass_val):
            if not self.is_val:
                self._image_summary(A.get_name(self.training_data),
                                    self.training_data,
                                    TRAIN_SUMMARY_COLLECTION)
            else:
                self._image_summary(A.get_name(self.val_data),
                                    self.val_data,
                                    VALID_SUMMARY_COLLECTION)

                # Legacy_TODO: think refactor this. Should not be here. Perhaps sensor
                # should not be part of processing layer, and has a val copy.
                self.done_first_pass_val = True

    def _image_summary(self, name, image_batch, collection):
            self.log("Do tensorboard summary on outputs {} of {}".format(
                name, self.name))

            A.summary.histogram(name,
                                image_batch,
                                collections=[collection])
            A.summary.image(name,
                            image_batch,
                            collections=[collection])

    def _forward(self, *args, **kwargs):
        """
        Generate placeholder or tensor variables to represent the the input
        data.
        """
        self.source.forward()

        if not self.is_val:
            if not self.done_first_pass:
                self.log("Forwarding data from training sensor ... ")

            if issubclass(type(self.source), sources.SupervisedSource):
                self.training_data, self.training_labels \
                    = self._forward_train()
                return [self.training_data, self.training_labels]
            else:
                self.training_data = self._forward_train()

                return [self.training_data]

        else:
            if not self.done_first_pass_val:
                self.log("Forwarding data from val sensor ... ")

            if issubclass(type(self.source), sources.SupervisedSource):
                self.val_data, self.val_labels = self._forward_val()

                return [self.val_data, self.val_labels]
            else:
                self.val_data = self._forward_val()

                return [self.val_data]


class ShuffleQueueSensor(six.with_metaclass(abc.ABCMeta, OldSensor)):
    """
    A `OldSensor` that holds a shuffle queue, which would pre-load
    `min_fraction_of_examples_in_queue` number of examples and sample batches
    from the queue. Preloading could reduce disk latency and doing data
    augmentation ahead of time, and load a relatively large number of samples
    could ensure that the random shuffling has good
    mixing properties.
    """

    def __init__(self, min_fraction_of_examples_in_queue=0.1, **kwargs):
        """
        Args:
            min_fraction_of_examples_in_queue: A fraction less than 1.
        """
        super(ShuffleQueueSensor, self).__init__(**kwargs)
        assert min_fraction_of_examples_in_queue > 0 and \
            min_fraction_of_examples_in_queue <= 1, \
            "min_fraction_of_examples_in_queue should be between 0 and 1"
        self.min_fraction_of_examples_in_queue \
            = min_fraction_of_examples_in_queue


class IntegratedSensor(ShuffleQueueSensor):
    """
    A concrete `OldSensor` uses Reader Op of tensorflow to read data directly as a
    tensor into a computational graph instead of being a placeholder as in
    `FeedSensor`.

    Optionally, it could also do data augmentation. It holds two
    `LinkedSystem`s, `training_jokers` and `val_jokers`, which do data
    processing on training datum and validation datum respectively.
    """
    def __init__(self, num_preprocess_threads=4,  **kwargs):
        super(IntegratedSensor, self).__init__(**kwargs)
        self.num_preprocess_threads = num_preprocess_threads

        # Keep two LinkedSystem to hold Jokers that may apply to training and
        # validation data.
        self.training_jokers = JokerSystem(name="training_joker")
        self.val_jokers = JokerSystem(name="val_joker")

    def _forward_train(self):
        # Legacy_TODO(Shuai): Handle the case where the source has no labels.
        self.training_jokers.forward(self.source.training_datum)
        augmented_training_datum = self.training_jokers.data
        min_queue_examples = int(self.source.num_train *
                                 self.min_fraction_of_examples_in_queue)

        batch_list = self._generate_image_and_label_batch(
            self.batch_size,
            augmented_training_datum,
            self.source.training_label,
            min_queue_examples,
            "train_data")
        training_data = batch_list[0]
        training_labels = batch_list[1:]

        return training_data, training_labels

    def _forward_val(self):
        # Legacy_TODO(Shuai): Handle the case where the source has no labels.
        self.val_jokers.forward(self.source.val_datum)
        processed_val_datum = self.val_jokers.data
        min_queue_examples = int(self.source.num_val *
                                 self.min_fraction_of_examples_in_queue)

        batch_list = self._generate_image_and_label_batch(
            self.val_batch_size,
            processed_val_datum,
            self.source.val_label,
            min_queue_examples,
            "val_data")
        val_data = batch_list[0]
        val_labels = batch_list[1:]

        return val_data, val_labels

    def attach(self, joker, to_val=False):
        """
        Attach a joker to a joker system. If `to_val` is True, attach to
        training joker system, otherwise to validation joker system.
        """
        if to_val:
            self.val_jokers.attach(joker)
        else:
            self.training_jokers.attach(joker)

    def _post_forward(self):
        super(IntegratedSensor, self)._post_forward()
        if self.do_summary:
            # Do image summary on raw images if we have done data augmentation.
            if not self.training_jokers.is_empty:
                self._raw_datum_summary(self.training_data.op.name + "_raw",
                                        self.source.training_datum,
                                        TRAIN_SUMMARY_COLLECTION)
        if self.do_summary_on_val and self.is_val and not self.val_jokers.is_empty:
            self._raw_datum_summary(self.val_data.op.name + "_raw",
                                    self.source.val_datum,
                                    VALID_SUMMARY_COLLECTION)

    def _raw_datum_summary(self, name, datum, collection):
        # Since image summary only takes image batches, we package each
        # image into a batch.
        tf.summary.image(name,
                         tf.expand_dims(datum, 0),
                         collections=[collection])

    def _generate_image_and_label_batch(
            self, batch_size, image, label, min_queue_examples, name):
        """Construct a queued batch of images and labels.

        Args:
            batch_size: An integer.
            image: 3-D Tensor of [IMAGE_SIZE, IMAGE_SIZE, 3] of type.float32.
            label: 1-D Tensor of type.int32 or a list of them.
            min_queue_examples: int32, minimum number of samples to retain
            in the queue that provides of batches of examples.

        Returns:
            batch_list: a list
                A list of batched tensors of passed in `image` and `label`. The
                order how they are passed in is preserved in the list.
        """
        input_list = [image]
        input_list.extend(label) if type(label) is list \
            else input_list.append(label)
        batch_list = tf.train.shuffle_batch(
            input_list,
            batch_size=batch_size,
            num_threads=self.num_preprocess_threads,
            capacity=min_queue_examples + 2 * self.num_preprocess_threads * batch_size,
            min_after_dequeue=min_queue_examples,
            name=name)

        for i, b in enumerate(batch_list):
            batch_list[i] = tf.squeeze(b)

        return batch_list


class FeedSensor(OldSensor):
    """
    Sense from a `FeedSource` to supply data to a `Kid`.
    """
    def _forward_train(self):
        return self._make_placeholder("train_data", self.batch_size)

    def _forward_val(self):
        return self._make_placeholder("val_data", self.val_batch_size)

    def _make_placeholder(self, name, batch_size):
        data_shape = self.source.shape
        data_shape.insert(0, batch_size)
        data = tf.placeholder(tf.float32, shape=data_shape, name=name)

        if issubclass(type(self.source), sources.SupervisedSource):
            label_shape = self.source.label_shape
            if len(label_shape) is 1:
                label_shape = [batch_size]
            else:
                label_shape.insert(0, batch_size)

            labels = tf.placeholder(tf.int32, shape=label_shape, name=name)

            return data, labels

        return data

    def fill_feed_dict(self):
        """Supply a batch of examples in form of feed dict. If `is_val` flag is
        true, it will return validation data, otherwise, it returns training
        data.

        A feed_dict takes the form of:
        feed_dict = {
            <placeholder>: <tensor of values to be passed for placeholder>,
            ....
        }

        Returns:
            feed_dict: The feed dictionary mapping from placeholders to values.
        """
        # Create the feed_dict for the placeholders filled with the next
        # `batch size ` examples.
        batch_size = self.val_batch_size if self.is_val else self.batch_size
        images_feed, labels_feed = self.source.get_batch(batch_size,
                                                         self.is_val)
        feed_dict = {
            self.data: images_feed,
            self.labels: labels_feed,
        }
        return feed_dict


class TorchSensor(OldSensor):
    def __init__(self, pin_memory=False, **kwargs):
        super(TorchSensor, self).__init__(**kwargs)
        self.pin_memory = pin_memory

    def _setup(self):
        self.source.setup()
        # Legacy_TODO: try use pin memory.
        self.loader = th.utils.data.DataLoader(self.source.dataset,
                                               batch_size=self.batch_size,
                                               shuffle=self.shuffle_train,
                                               pin_memory=self.pin_memory,
                                               num_workers=8)
        self.val_loader = th.utils.data.DataLoader(self.source.val_dataset,
                                                   batch_size=self.val_batch_size,
                                                   pin_memory=self.pin_memory,
                                                   num_workers=8)
        self.iter = self.loader.__iter__()
        self.val_iter = self.val_loader.__iter__()

    def next(self):
        if self.is_val:
            try:
                return self.val_iter.next()
            except StopIteration:
                self.val_iter = self.val_loader.__iter__()
                return self.val_iter.next()
        else:
            try:
                return self.iter.next()
            except StopIteration:
                self.iter = self.loader.__iter__()
                A.inc_epoch()
                return self.iter.next()

    def _forward_train(self):
        ret = [th.autograd.Variable(t.cuda() if A.use_cuda() else t) for t in self.next()]
        A.cache_tensor_auto_scope(ret[0], "data")
        A.cache_tensor_auto_scope(ret[1], "labels")
        return ret

    def _forward_val(self):
        ret = [th.autograd.Variable(t.cuda() if A.use_cuda() else t) for t in self.next()]
        A.cache_tensor_auto_scope(ret[0], "val_data")
        A.cache_tensor_auto_scope(ret[1], "val_labels")
        return ret


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



from .. import backend as A


regularizers = {}


def compute(reg_name, **kwargs):
    reg = regularizers[reg_name]

    # Make sure required fields exist.
    for p in reg.required_fields:
        if p not in kwargs:
            raise KeyError("Required field {} not found in the provided"
                           " initialization parameters, `init_para`. Perhaps"
                           " you have some typos.".format(p))

    return reg.func(**kwargs)


class RegularizerRegistry(object):
    def __init__(self,
                 name,
                 func,
                 required_fields=(),
                 msg=None):
        self.name = name
        self.func = func
        self.required_fields = required_fields
        self.msg = msg
        regularizers[name] = self


def l2_regularizer(var, scale, name=None):
    return A.mul(A.nn.l2_loss(var), scale, name=name)


def l1_regularizer(var, scale, name=None):
    return A.mul(A.nn.l1_loss(var), scale, name=name)


RegularizerRegistry("l2",
                    l2_regularizer,
                    required_fields=("var", "scale"),
                    msg="l2 regularization")
RegularizerRegistry("l1",
                    l1_regularizer,
                    required_fields=("var", "scale"),
                    msg="l1 regularization")



"""
Signals propagated in nature are all abstracted as a source.  Source is an
abstraction of measurements collections, e.g., a dataset.

As an example, saying in supervised setting, a source is a block that takes no
inputs (since it is a source), and outputs data. A concrete example could be
the source for the MNIST dataset::

    source = MNISTSource(name="MNIST",
                         work_dir=AKID_DATA_PATH + '/mnist')

The above code creates a source for MNIST. It provides data by indices through
`get` method. For example::

    source.get([0, 1, 2])

would return a list of Tensor `[images, labels]`, where the first is a Tensor
of which the first dimension is 3 --- each stores an image; and the second is a
Tensor that stores three labels.

It could be used standalone, or passed to a `Sensor`.

A source has multiple modes, i.e., "train", "val", or "test". Different models
provide different data. For example, "train" model provides training data. To
change the model, call::

    source.set_mode("val")
    source.setup()

`setup` should be called after mode change. It sets up things like data
prefetching threads.

Developer Note
================

A top level abstract class `Source` implements basic semantics of a natural
source. Other abstract classes keep implementing more concrete
sources. Abstract `Source` s need to be inherited and abstract methods
implemented before it could be used.
"""
from __future__ import absolute_import
from __future__ import print_function
import abc
import sys
import inspect
import os
import six.moves.urllib.request, six.moves.urllib.parse, six.moves.urllib.error
import tarfile

import numpy as np
import tensorflow as tf
from torch.utils.data.dataloader import default_collate

from .blocks import DataBlock, FlowBlock
from .. import backend as A
from .common import AKID_DATA_PATH
import six
from six.moves import range


# Basic model parameters as external flags.
flags = tf.app.flags
FLAGS = flags.FLAGS
flags.DEFINE_boolean('fake_data', False, 'If true, uses fake data '
                     'for unit testing.')


class Source(DataBlock):
    """
    A source that holds datasets, which include training, validation, and test
    datasets.

    Subclasses should implement their own `_get` method normally (though a
    default `_get` method is implemented for PyTorch for now).
    """
    NAME = "Source"

    def __init__(self, work_dir=AKID_DATA_PATH + "/data", *args, **kwargs):
        """
        By default, the source is in `train` mode.
        Args:
            work_dir: str
                The directory to store data.
        """
        super(Source, self).__init__(*args, **kwargs)
        self.mode = A.Mode.TRAIN
        self.work_dir = work_dir

    @abc.abstractproperty
    def data(self):
        """
        The method should return a reference that points to the whole dataset
        in its current mode.

        It could be an iterator when the dataset is large, or an array when the
        dataset is small.
        """
        raise NotImplementedError("Each source needs to implement this"
                                  " method to provide an interface to offer"
                                  " data!")

    @abc.abstractproperty
    def size(self):
        """
        The method should return the size of the data source.
        """
        pass

    def set_mode(self, mode):
        A.check_mode(mode)
        self.mode = mode
        # Each time `set_mode` is called, we need re-set up the source. It is
        # easy to miss it, so we make it explicit.
        self.is_setup = False

    def get(self, indices):
        if not self.is_setup:
            raise ValueError("Source is not set up yet.")

        return self._get(indices)

    def _get(self, indices):
        """
        This is a default `get` method to work with the `DataSet` class of
        PyTorch. A customized source that does not use `DataSet` in PyTorch
        should implement a `get` method of its own.
        """
        # Only work in torch for now.
        sample_batch = default_collate([self.data[i] for i in indices])
        return sample_batch


class OldSource(six.with_metaclass(abc.ABCMeta, FlowBlock)):
    """
    An abstract class to model data source from the world.

    The forms in which certain data arrived varies as many ways as one could
    imagine. This class abstracts this complexity and provides a uniform
    interface for subsequent processing.

    According to how tensorflow supplies data, two ways existed to
    correspondingly, `FeedSource` and `TFSource`.
    """

    def __init__(self,
                 url=None,
                 work_dir="data",
                 validation_rate=None,
                 **kwargs):
        """
        Args:
            url: str
                Uniform Resource Locator. It could point to a file or web
                address, where this source should fetch data from.
            work_dir: str
                Working directory of the `OldSource`. It may puts temporal files,
                such as downloaded dataset and so on. By default, use a folder
                named data under root directory.
        validation_rate: a percentage
            The proportion of training data to be used as validation set. If
            None, all training data will be used for training.
        """
        super(OldSource, self).__init__(**kwargs)
        self.url = url
        self.work_dir = work_dir

        if validation_rate:
            assert validation_rate >= 0 and validation_rate < 1,\
                "The percentage used for validation must be between 0-1."
            self.validation_rate = validation_rate
        else:
            self.validation_rate = 0

        if self.url:
            # Only download if url is specified. Ideally, the automatic
            # download is good, however, for fast prototype purpose, normally
            # people do not bother to automate the process, so it is better to
            # support this option.
            self._get_raw_data_if_not_yet()

    def _get_raw_data_if_not_yet(self):
        """
        Download anything pointed by `self.url` and save it under
        `self.work_dir`. If the file is a tarball, it will be extracted.
        """
        if not os.path.exists(self.work_dir):
            os.makedirs(self.work_dir)

        filename = self.url.split('/')[-1]
        filepath = os.path.join(self.work_dir, filename)
        if not os.path.exists(filepath):
            def _progress(count, block_size, total_size):
                sys.stdout.write(
                    '\r>> Downloading %s %.1f%%' %
                    (filename,
                        float(count * block_size) /
                        float(total_size) * 100.0))
                sys.stdout.flush()
            filepath, _ = six.moves.urllib.request.urlretrieve(
                self.url, filepath, reporthook=_progress)
            print()
            statinfo = os.stat(filepath)
            print(('Succesfully downloaded',
                  filename, statinfo.st_size, 'bytes.'))
            # Extract if a tarball.
            suffixes = filename[filename.find('.')+1:]
            if suffixes == 'tar.gz':
                tarfile.open(filepath, 'r:gz').extractall(self.work_dir)

    def data(self):
        """
        This is the first place a multiple outputs mechanism is
        needed. However, a temporary solution is used now, so this method is
        actually not used yet.
        """
        assert False, "Program should not reach here."

    @property
    def training_datum(self):
        """
        An optional property to provide a datum of particular training data
        source. It is not made an abstract method given some sub-classes are
        only able to provide shape information, aka `FeedSource`.
        """
        raise NotImplementedError("The property `training_datum` is not"
                                  " implemented!")
        sys.exit()

    @property
    def val_datum(self):
        """
        An optional property to provide a validation datum of particular data
        source. It is not made abstract for the same reason with
        `training_data`.
        """
        raise NotImplementedError("The property `val_datum` is not"
                                  "implemented!")
        sys.exit()

    @property
    def shape(self):
        """
        An optional property to provide shape of particular training data
        source. It is not made an abstract method since if some `OldSource` class
        returns data tensor directly, it already contains shape information.
        """
        raise NotImplementedError("The property `shape` is not implemented!")
        sys.exit()


class SupervisedSource(OldSource):
    """
    An abstract class to model supervised data source.
    """
    @property
    def label_shape(self):
        """
        An optional property to provide shape of particular label of training
        data source. It is not made an abstract method for the same reason with
        `shape` of `OldSource`.
        """
        raise NotImplementedError("The property `label_shape` is not"
                                  " implemented!")
        sys.exit()

    @property
    def training_label(self):
        """
        An optional property to provide a training label of particular data
        source. It is not made an abstract method given some sub-classes are
        only able to provide shape information.

        If a sub-class decides to use this property, the setup of tensor
        `labels` should be in `_forward`.
        """
        raise NotImplementedError("The property `training_label` is not"
                                  " implemented!")
        sys.exit()

    @property
    def val_label(self):
        """
        An optional property to provide a validation label of particular data
        source. It is not made abstract for the reason same with property
        `training_labels`.

        If a sub-class decides to use this property, the setup of tensor
        `labels` should be in `_forward`.
        """
        raise NotImplementedError("The property `val_label` is not"
                                  " implemented!")
        sys.exit()


class StaticSource(OldSource):
    """
    An abstract class to model static dataset partitioned into training data
    and test data.
    """
    def __init__(self, num_train, num_val, **kwargs):
        super(StaticSource, self).__init__(**kwargs)
        self.num_train = num_train
        self.num_val = num_val

    @property
    # Legacy_TODO(Shuai): This property should be used to deal with the case batch
    # cannot divide the number of test data. It should be updated
    # accordingly. For instance, for InMemoryFeedSource, the information comes
    # from `_epochs_completed` of `DataSet`.
    def epochs_completed(self):
        return self._epochs_completed


class FeedSource(OldSource):
    """
    An abstract class that supplies data in form of numpy.array.

    It does not create any tensor, and only plays the role to supply meta
    information of data to further `FeedSensor`, and of course supply actual
    data.

    Every concrete sub-class should implement a `get_batch` method to actually
    supply data, in the shape desired.
    """
    def __init__(self, center=False, scale=False, **kwargs):
        super(FeedSource, self).__init__(**kwargs)
        self.center = center
        self.scale = scale

    @abc.abstractmethod
    def get_batch(self, num, get_val):
        """
        Return `num` of datum, either in one numpy.array, or a tuple of
        numpy.array.

        Args:
            get_val: Boolean
                If True, get from validation samples or training samples.
        """
        raise NotImplementedError("Each sub `FeedSource` needs to implement"
                                  " this method to actually supply data.")
        sys.exit()


class InMemoryFeedSource(StaticSource, FeedSource):
    """
    An abstract class to load all data into memory.

    This class holds a private member `_data_sets` to hold all data.
    """
    def _forward(self):
        """
        Call `_load` to load datasets into memory.
        """
        # Read the whole dateset into memory.
        self.data_sets = self._load()

    def get_batch(self, num, get_val):
        if get_val:
            return self.data_sets.test.next_batch(num)
        else:
            return self.data_sets.training.next_batch(num)

    def get_all(self, train):
        """
        Get all samples in the source.

        Args:
            train: Boolean
                Get from training samples or test samples.

        Returns:
            dataset: datasets.Dataset
        """
        if train:
            return self.data_sets.training
        else:
            return self.data_sets.test

    @abc.abstractmethod
    def _load(self):
        """
        Load data into memory.

        Returns:
            datasets: dataset.DataSets
        """
        raise NotImplementedError("Each sub `InMemorySouce` needs to implement"
                                  " this method to load data.")
        sys.exit()


class TFSource(StaticSource):
    """
    An abstract class that uses Reader Op of tensorflow to supply data.

    It takes no inputs, and outputs four tensors --- training data, validation
    data, training labels, validation labels. Due to its clear semantics, each
    of those tensors is a property (instead of a list of tensors).

    Since normally usage of `TFSource` accompanies with data augmentation, and
    the way data augmentation works at the granularity of one sample, so
    `_forward` of `TFSource` should initialized `training_datum` and `val_datum`
    to a `tf.Tensor` that returns by some Reader Op of tensorflow. The data
    provided by this class of source has necessary information associated with
    the tensor variable, and could be used directly in the further pipeline.

    Note the optional properties of `OldSource`, is made abstract, consequently
    mandatory.
    """
    @abc.abstractmethod
    def _forward(self):
        self._read()
        """
        TFSource uses Reader Ops of Tensorflow to read data. So any sub-classes
        of `TFSource` should implement it to actually read data. If it is
        combined with `SupervisedSource`, then setup of `labels` should also be
        put here.
        """
        raise NotImplementedError("Each sub-class of TFSource needs to"
                                  " implement this method to read data!")
        sys.exit()

    @abc.abstractproperty
    def training_datum(self):
        """
        An abstract property to enforce any sub-classes to provide training
        data in form of tf.Tensor.
        """
        raise NotImplementedError("Each sub-class of TFSource needs to"
                                  " implement this method to provide a"
                                  " training datum!")
        sys.exit()

    @abc.abstractproperty
    def val_datum(self):
        """
        An abstract property to enforce any sub-classes to provide training
        data in form of tf.Tensor.
        """
        raise NotImplementedError("Each sub-class of TFSource needs to"
                                  " implement this method to provide"
                                  " a validation datum!")
        sys.exit()

    def _float_feature(self, value):
        """
        Helper method for construct Feature protobuf for tfrecord.

        `_int_feature`, `_bytes_feature` are similar methods.
        Args:
            value: list
                A list that holds float values to store.
        """
        return tf.train.Feature(float_list=tf.train.FloatList(value=value))

    def _int_feature(self, value):
        """
        See `_float_feature`.
        """
        return tf.train.Feature(int64_list=tf.train.Int64List(value=value))

    def _bytes_feature(self, value):
        """
        See `_float_feature`.
        """
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


class ClassificationTFSource(TFSource, SupervisedSource):
    """
    An abstract class supplies data using tfrecords for classification problem.

    It further makes concrete most of the abstract methods of its super
    classes.
    """
    @property
    def training_datum(self):
        return self._training_datum

    @property
    def val_datum(self):
        return self._val_datum

    @property
    def training_label(self):
        return self._training_label

    @property
    def val_label(self):
        return self._val_label

    def _convert_to_tf(self, images, labels, name):
        """
        Take a numpy array of images and corresponding labels and convert it to
        tfrecord format.

        Args:
            images: numpy array
                images of shape of shape [N, H, w, C].
            labels: numpy array of shape [N] or list
                corresponding labels
            name: a str
                The output tfrecord file will be named `name.tfrecords`.
        """
        num_examples = labels.shape[0]
        if images.shape[0] != num_examples:
            raise ValueError("Images size %d does not match label size %d." %
                             (images.shape[0], num_examples))
        row = images.shape[1]
        col = images.shape[2]
        depth = images.shape[3]

        filename = os.path.join(self.work_dir, name + '.tfrecords')
        print(('Writing', filename))
        writer = tf.python_io.TFRecordWriter(filename)
        for index in range(num_examples):
            image_raw = np.reshape(images[index], -1).tolist()
            example = tf.train.Example(features=tf.train.Features(
                feature={
                    'height': self._int_feature([row]),
                    'width': self._int_feature([col]),
                    'depth': self._int_feature([depth]),
                    'label': self._int_feature([int(labels[index])]),
                    'image_raw': self._float_feature(image_raw)}))
            writer.write(example.SerializeToString())
        writer.close()


# TODO:
# The following lines are for a cleaner namespace for akid, since those
# modules, classes are introduced to akid by import *. However, if I do this,
# sphinx cannot find those classes as well. I think I should manual type in
# what classes I want in __all__. This change is cascaded --- all inspect code
# are deleted in all modules; no TODO in those places anymore..
# __all__ = [name for name, x in locals().items() if
#            not inspect.ismodule(x) and not inspect.isabstract(x)]



from __future__ import absolute_import
from akid import AKID_DATA_PATH
from akid import FeedSensor
from akid import Kid
from akid import MomentumKongFu
from akid import MNISTFeedSource

from akid.models.brains import NewMnistTfTutorialNet
from akid import LearningRateScheme
from akid.utils.test import TestFactory
from akid import backend as A


def update_lr(kid):
    kid.kongfu.set_lr(kid.kongfu.get_lr() * 0.95)


def setup():
    brain = NewMnistTfTutorialNet(name="mnist-tf-tutorial-net")
    sensor = TestFactory.get_test_sensor()
    kid = Kid(
        sensor,
        brain,
        MomentumKongFu(
            lr=0.01,
            # lr_scheme here does not have effect, only left for reference.
            lr_scheme={
                "name": LearningRateScheme.exp_decay,
                "base_lr": 0.01,
                "decay_rate": 0.95,
                "num_batches_per_epoch": 468,
                "decay_epoch_num": 1},
            momentum=0.9),
        max_steps=20000)

    kid.hooks.on_epoch_end.append(update_lr)

    kid.setup()

    return kid


if __name__ == "__main__":
    kid = setup()
    kid.practice()



from akid import backend as A

@A.cache_name_if_exist
def hinge_loss(y, t, weights=None, name=None):
    """
    Given 1D, or 2D Tensor y, t, calculate the (weighted) hinge loss as follows
    for each element in the tensor, and average them. That means, that it
    supports binary classification, and multi-label binary classification.

    .. math::
        l(y, t) = weights * \max{0, 1 - t * y}

    TODO: implement margin in the loss.
    """
    loss = 1 - t * y
    loss[loss < 0] = 0
    if weights is not None:
        loss *= weights
    loss = A.mean(loss)
    return loss


@A.cache_name_if_exist
def tripod_hinge_loss(y, t, weights=None, signal_balancing=False, name=None):
    """
    Given 1D, or 2D Tensor y, t, calculate the (weighted) hinge loss as follows
    for each element in the tensor, and average them. That means, that it
    supports binary classification, and multi-label binary classification.

    if t = -1, or 1
    .. math::
        l(y, t) = weights * \max{0, 1 - t * y}

    if t = 0
    .. math::
        l(y, t) = weights * |y|

    TODO: implement margin in the loss.
    Args:
        signal_balancing: bool
            If True, will average the loss of different labels first, and then
            average the mean loss of different labels. It puts equal weights on
            different label, thus balancing the different label values.
    """
    middle_idx = A.cast(t, A.uint8) == 0
    middle_score = y[middle_idx]

    end_idx = ~middle_idx
    end_score = y[end_idx]
    end_label = t[end_idx]

    end_score_loss = 1 - end_label * end_score
    end_score_loss[end_score_loss < 0] = 0
    middle_score_loss = A.abs(middle_score)

    # TODO: Deal with weight later.
    # if weights is not None:
    #     loss *= weights

    if signal_balancing:
        loss = (A.mean(end_score_loss) + A.mean(middle_score_loss)) / 2
    else:
        loss = A.cat([end_score_loss, middle_score_loss])
        loss = A.mean(loss)
    return loss


@A.cache_name_if_exist
def hinge_ranking_loss(x1, x2, margin, hard_sample_mining=False, name=None):
    """
    Hinge loss adopted for ranking. To understand the loss, recall that hinge
    loss is an almost stair-case style approximation to 0-1 loss, and 0-1 loss
    can be understood as a relationship between two states. For ranking, we
    need more than one state to represent the ranking order. More details can
    be found at *Ranking with Ordered Weighted Pairwise Classification*, the
    paper that motivates the loss.

    Currently in this implementation, only ranking order between positive and
    negative samples are characterized. The loss is of the formula

    .. math::
        l(x1, x2) = \max{0, margin - (x1 - x2)}

    where x1 is the positive sample, and x2 is the negative sample.

    Intuitively, compared with hinge loss, it simply enforces a margin between
    x1 and x2, instead of dictating that the score should be larger than an
    absolute value as in the hinge loss, :meth:`akid.nn.losses.hinge_loss`.

    Args:
        hard_sample_mining: bool
            Whether to use hard sample mining in the loss. It would use the
            pair (x1, x2), where x1 < x2.
    """
    delta = (x1 - x2)
    if hard_sample_mining:
        delta = delta[delta - margin < 0]
    loss = margin - delta
    loss[loss < 0] = 0
    loss = A.mean(loss)
    return loss



from __future__ import absolute_import
import sys

def currentframe():
    """Return the frame object for the caller's stack frame."""
    try:
        raise Exception
    except:
        return sys.exc_info()[2].tb_frame.f_back


def is_tuple_or_list(v):
    if isinstance(v, tuple) or isinstance(v, list):
        return True
    else:
        return False



"""
See the argument parser's description below for introduction and usage.

# Mechanism
# #########################################################################
Given the available GPU numbers, a semaphore is created to control access to
GPUs. A lock is created to control access to the mask to indicator which GPU is
available. After a process has modified the gpu mask, it releases the lock
immediately, so other process could access it. But the semaphore is still not
release, since it is used to control access to the actual GPU. A training
instance will be launched in a subshell using the GPU acquired. The semaphore
is only released after the training has finished.
"""
from __future__ import absolute_import
from __future__ import print_function
import multiprocessing
import importlib
import argparse

import tensorflow as tf
import pycuda.autoinit
import pycuda.driver as cuda
from six.moves import range


def spawn(s, l, gpu_mask, **kwargs):
    with s:
        # Look up a GPU and mark it used.
        # A lock is unnecessary for manager list, but in order to let the
        # printed information print right, a lock is used to control access to
        # stdout.
        with l:
            for idx, avail in enumerate(gpu_mask):
                if avail == 1:
                    print(("GPU mask {}.".format(gpu_mask)))
                    print(("Using GPU {}.".format(idx)))
                    gpu_mask[idx] = 0
                    break

        # Train using the acquired GPU.
        with tf.device("/gpu:{}".format(idx)):
            # Built the graph here.
            kid = trainee.setup(tf.get_default_graph(), **kwargs)

        # Actually call the code that run a session.
        kid.practice()

        # Release the GPU.
        with l:
            print(("Released GPU {}.".format(idx)))
            gpu_mask[idx] = 1


# Setup argument parser.
parser = argparse.ArgumentParser(
    description="""
Description:
# #########################################################################
A template for tuning parameters using multiple GPUs. Different trainees will
be put on different GPUs.

Usage:
# #########################################################################
A trainee is a script with a function `setup` that would set up and return a
Kid, but not train it, which is to say, call the `practice` method.

The `setup` function should expose the network parameter to be tuned as formal
parameters.

Currently, supported parameters are:

    * learning rate

See maxout.py in the same folder for an example.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument("TRAINEE_FILENAME",
                    type=str,
                    help="the path to the table to be drawn. The file has to"
                    " be in the same folder for now.")
parser.add_argument("-lr", "--learning-rates",
                    type=float,
                    nargs='+',
                    help="The list of learning rates to try.")
arguments = parser.parse_args()

# Gather trainee info.
# #########################################################################
# Strip the file suffix
trainee_name = arguments.TRAINEE_FILENAME.split('.')[0]
trainee = importlib.import_module(trainee_name)

lr_list = arguments.learning_rates

# Set up data structures.
# #########################################################################
manager = multiprocessing.Manager()
gpu_num = cuda.Device.count()
gpu_mask = manager.list([1] * gpu_num)

# Logistics
# #########################################################################
s = multiprocessing.Semaphore(len(gpu_mask))
l = multiprocessing.Lock()
process_pool = []

# Start tuning.
# #########################################################################
for i in range(len(lr_list)):
    p = multiprocessing.Process(target=spawn,
                                args=(s, l, gpu_mask),
                                kwargs={"lr": lr_list[i]})
    process_pool.append(p)
    p.start()

# Wait for all processes to finish.
for p in process_pool:
    p.join()
